from pathlib import Path

import numpy as np
import pytest

from pyaerocom.io.ghost.reader import ReadGhost


@pytest.fixture(scope="module")
def ghost_eea_daily():
    return ReadGhost("G.EEA.daily.Subset")


@pytest.fixture(scope="module")
def ghost_eea_hourly():
    return ReadGhost("G.EEA.hourly.Subset")


class TestReadGhost:
    PROVIDES_VARIABLES = [
        "concpm10",
        "concpm10al",
        "concpm10as",
        "concpm25",
        "concpm1",
        "conccl",
        "concso4",
        "vmrco",
        "vmrno",
        "vmrno2",
        "vmro3",
        "vmrso2",
        "concno3",
        "concnh4",
        "concco",
        "concno",
        "concno2",
        "conco3",
        "concso2",
    ]

    INVDICT = {
        "pm10": "concpm10",
        "pm10al": "concpm10al",
        "pm10as": "concpm10as",
        "pm2p5": "concpm25",
        "pm1": "concpm1",
        "sconccl": "conccl",
        "sconcso4": "concso4",
        "sconcco": "vmrco",
        "sconcno": "vmrno",
        "sconcno2": "vmrno2",
        "sconco3": "vmro3",
        "sconcso2": "vmrso2",
        "sconcno3": "concno3",
        "sconcnh4": "concnh4",
    }

    DEFAULT_VAR = "conco3"
    DEFAULT_SITE = None

    fixture_names = ("ghost_eea_daily", "ghost_eea_hourly")

    # cf.: https://hackebrot.github.io/pytest-tricks/fixtures_as_class_attributes/
    @pytest.fixture(autouse=True)
    def auto_injector_fixture(self, request):
        for name in self.fixture_names:
            setattr(self, name, request.getfixturevalue(name))

    @property
    def default_reader(self):
        """Just a convenience thing used in tests below that would be redundant
        if applied to daily and hourly"""
        return self.get_reader("ghost_eea_daily")

    def get_reader(self, fixture_name):
        return getattr(self, fixture_name)

    def test_meta_keys(self):
        assert len(self.default_reader.META_KEYS) == 142

    @pytest.mark.parametrize("fixture_name", ["ghost_eea_daily", "ghost_eea_hourly"])
    def test_PROVIDES_VARIABLES(self, fixture_name):
        reader = self.get_reader(fixture_name)
        assert len(reader.PROVIDES_VARIABLES) == len(self.PROVIDES_VARIABLES)
        assert all([x in self.PROVIDES_VARIABLES for x in reader.PROVIDES_VARIABLES])

    def test_var_names_data_inv(self):
        invdict = self.default_reader.var_names_data_inv
        assert isinstance(invdict, dict)
        for key, val in invdict.items():
            assert self.INVDICT[key] == val

    @pytest.mark.parametrize(
        "fixture_name,vars_to_read,pattern,filenum,lastfilename",
        [
            ("ghost_eea_daily", "vmro3", None, 1, "sconco3_201810.nc"),
            ("ghost_eea_hourly", "vmro3", None, 1, "sconco3_201810.nc"),
            ("ghost_eea_daily", "concpm10", None, 3, "pm10_201912.nc"),
            ("ghost_eea_daily", "vmro3", "*201810.nc", 1, "sconco3_201810.nc"),
        ],
    )
    def test_get_file_list(self, fixture_name, vars_to_read, pattern, filenum, lastfilename):
        files = self.get_reader(fixture_name).get_file_list(
            vars_to_read=vars_to_read, pattern=pattern
        )

        assert len(files) == filenum
        assert Path(files[-1]).name == lastfilename

    def test__ts_type_from_data_dir(self):
        assert self.get_reader("ghost_eea_daily")._ts_type_from_data_dir() == "daily"

    @pytest.mark.parametrize(
        "fixture_name,val", [("ghost_eea_daily", "daily"), ("ghost_eea_hourly", "hourly")]
    )
    def test_TS_TYPE(self, fixture_name, val):
        assert self.get_reader(fixture_name).TS_TYPE == val

    def test_get_meta_filename(self):
        import pandas as pd

        meta = self.default_reader.get_meta_filename("pm10_201910.nc")
        per = pd.Period(freq="M", year=2019, month=10)
        desired = dict(var_name="pm10", start=per.start_time, stop=per.end_time)
        for key, val in meta.items():
            assert desired[key] == val

    def test__eval_flags_slice(self):
        import xarray as xr

        reader = self.default_reader
        file = reader.files[-1]
        assert Path(file).name == "sconco3_201810.nc"
        ds = xr.load_dataset(file)

        flagvar = "qa"
        numvalid = 3
        shape = (1, 3)

        assert "sconco3" in ds
        assert ds["sconco3"].shape == shape
        assert flagvar in reader.FLAG_DIMNAMES

        flagvar_dimname = reader.FLAG_DIMNAMES[flagvar]

        assert flagvar_dimname in ds.dims

        flags = ds[flagvar]

        invalidate = reader.DEFAULT_FLAGS_INVALID[flagvar]

        slice_dim = flags.dims.index(flagvar_dimname)
        valid = np.apply_along_axis(reader._eval_flags_slice, slice_dim, flags.values, invalidate)
        assert valid.ndim == 2
        assert valid.shape == shape
        assert valid.sum() == numvalid

    @pytest.mark.parametrize(
        "fixture_name,statnum,first_stat_name",
        [
            ("ghost_eea_daily", 1, "Bleak House"),
        ],
    )
    def test_read_file(self, fixture_name, statnum, first_stat_name):
        reader = self.get_reader(fixture_name)
        data = reader.read_file(reader.files[-1])
        assert isinstance(data, list)
        assert len(data) == statnum
        first_stat = data[0]
        assert isinstance(first_stat, dict)
        assert first_stat["meta"]["station_name"] == first_stat_name



from __future__ import annotations

import pytest

from pyaerocom.io.ebas_file_index import EbasFileIndex, EbasSQLRequest


def test_EbasSQLRequest___init__():
    EbasSQLRequest()


@pytest.mark.parametrize(
    "var,output",
    [
        ("bla", "('bla')"),
        (("bla", "blub"), "('bla', 'blub')"),
        (["bla", "blub"], "('bla', 'blub')"),
    ],
)
def test_EbasSQLRequest__var2sql(var: str | tuple[str] | list[str], output: str):
    req = EbasSQLRequest()
    assert req._var2sql(var) == output


def test_EbasSQLRequest__var2sql_error():
    with pytest.raises(ValueError) as e:
        EbasSQLRequest()._var2sql({})
    assert str(e.value) == "Invalid value..."


@pytest.mark.parametrize(
    "kwargs,output",
    [
        (
            {},
            "select distinct filename from variable join station on station.station_code=variable.station_code and not exists (select * from characteristic where var_id=variable.var_id and ct_type='Fraction');",
        ),
        (
            {"distinct": False},
            "select filename from variable join station on station.station_code=variable.station_code and not exists (select * from characteristic where var_id=variable.var_id and ct_type='Fraction');",
        ),
    ],
)
def test_EbasSQLRequest_make_file_query_str(kwargs: dict, output: str):
    req = EbasSQLRequest()
    assert req.make_file_query_str(**kwargs) == output


@pytest.mark.parametrize(
    "kwargs,output",
    [
        (
            {
                "altitude_range": [10, 21],
                "lon_range": [10, 20],
                "lat_range": [10, 20],
                "start_date": "2010-01-21",
                "stop_date": "2010-01-24",
                "statistics": ["arithmetic_mean", "median"],
                "datalevel": 2,
            },
            (
                "select distinct filename from variable join station on "
                "station.station_code=variable.station_code where "
                "station_altitude>10 and station_altitude<21 and "
                "station_longitude>10 and station_longitude<20 and "
                "station_latitude>10 and station_latitude<20 and "
                "first_end < '2010-01-24' and last_start > '2010-01-21' and "
                "statistics in ('arithmetic_mean', 'median') and datalevel=2;"
            ),
        ),
        (
            {},
            "select distinct filename from variable join station on station.station_code=variable.station_code;",
        ),
        (
            {"distinct": False},
            "select filename from variable join station on station.station_code=variable.station_code;",
        ),
        (
            {"what": ("filename", "station_code", "bla")},
            "select distinct filename,station_code,bla from variable join station on station.station_code=variable.station_code;",
        ),
    ],
)
def test_EbasSQLRequest_make_query_str(kwargs: dict, output: str):
    req = EbasSQLRequest()
    assert req.make_query_str(**kwargs) == output


def test_EbasSQLRequest___str__():
    assert isinstance(str(EbasSQLRequest()), str)


def test_EbasFileIndex___init__():
    EbasFileIndex()


def test_EbasFileIndex_database_getter():
    with pytest.raises(AttributeError) as e:
        EbasFileIndex().database
    error = "EBAS SQLite database file could not be located but is needed in EbasFileIndex class"
    assert str(e.value) == error


def test_EbasFileIndex_ALL_STATION_NAMES(ebas: EbasFileIndex):
    assert isinstance(ebas.ALL_STATION_NAMES, list)


def test_EbasFileIndex_ALL_STATION_CODES(ebas: EbasFileIndex):
    assert isinstance(ebas.ALL_STATION_CODES, list)


def test_EbasFileIndex_ALL_STATISTICS_PARAMS(ebas: EbasFileIndex):
    assert isinstance(ebas.ALL_STATISTICS_PARAMS, list)


def test_EbasFileIndex_ALL_VARIABLES(ebas: EbasFileIndex):
    assert isinstance(ebas.ALL_VARIABLES, list)


def test_EbasFileIndex_ALL_MATRICES(ebas: EbasFileIndex):
    assert isinstance(ebas.ALL_MATRICES, list)


def test_EbasFileIndex_ALL_INSTRUMENTS(ebas: EbasFileIndex):
    assert isinstance(ebas.ALL_INSTRUMENTS, list)


def test_EbasFileIndex_get_table_names(ebas: EbasFileIndex):
    # assert ebas.get_table_names() == ["station", "variable"]
    assert ebas.get_table_names() == ["station", "variable", "characteristic"]


table_comulns = dict(
    station=[
        "station_code",
        "platform_code",
        "station_name",
        "station_wdca_id",
        "station_gaw_name",
        "station_gaw_id",
        "station_airs_id",
        "station_other_ids",
        "station_state_code",
        "station_landuse",
        "station_setting",
        "station_gaw_type",
        "station_wmo_region",
        "station_latitude",
        "station_longitude",
        "station_altitude",
    ],
    variable=[
        "var_id",
        "station_code",
        "matrix",
        "comp_name",
        "statistics",
        "instr_type",
        "instr_ref",
        "method",
        "first_start",
        "first_end",
        "last_start",
        "last_end",
        "revdate",
        "period",
        "resolution",
        "datalevel",
        "filename",
        "vnum",
    ],
    characteristic=[
        "var_id",
        "ct_type",
        "datatype",
        "val_int",
        "val_dbl",
        "val_chr",
    ],
)


@pytest.mark.parametrize("table,column_names", table_comulns.items())
def test_EbasFileIndex_get_column_names(ebas: EbasFileIndex, table: str, column_names: list[str]):
    assert ebas.get_table_columns(table) == column_names



from __future__ import annotations

import logging

# from datetime import datetime, timezone, tzinfo
from pathlib import Path
from types import SimpleNamespace
from typing import Protocol

import pandas as pd

logger = logging.getLogger(__name__)


"""CAMS2-50 domain boundaries"""
CAMS2_50_DOMAIN = SimpleNamespace(
    lat=(30, 72),  # °N
    lon=(-25, 45),  # °E
)


class Domain(Protocol):
    """domain for static type checking"""

    lat: tuple[float, float]
    lon: tuple[float, float]


def in_domain(df: pd.DataFrame, *, domain: Domain) -> pd.Series:
    """check if rows are inside the domain"""
    return df["lon"].between(*domain.lon) & df["lat"].between(*domain.lat)


def add_time(df: pd.DataFrame) -> pd.DataFrame:
    """combine year/month/day/hour into a singre datetime column"""
    # dt = lambda row: datetime(row.Y, row.M, row.D, row.H, tzinfo=timezone.utc) # what the hell is this?

    df = df.rename(columns={"Y": "year", "M": "month", "D": "day", "H": "hour"})
    time = pd.to_datetime(df[["year", "month", "day", "hour"]], utc=True)

    return df.assign(time=time).drop(["year", "month", "day", "hour"], axis="columns")


def conc_units(df: pd.DataFrame) -> pd.DataFrame:
    """convert kg/m3 to μg/m3"""
    return df.assign(conc=df.conc * 1e9)


def poll_names(df: pd.DataFrame) -> pd.DataFrame:
    """rename pollutant names"""
    poll = df.poll.str.upper().str.replace("PM2P5", "PM25", regex=False)
    return df.assign(poll=poll)


def read_csv(
    path: str | Path, *, domain: Domain = CAMS2_50_DOMAIN, polls: list[str] = None
) -> pd.DataFrame:
    df = pd.read_csv(
        path,
        sep=";",
        header=0,
        names="station lat lon alt poll Y M D H _ conc".split(),
        usecols=lambda x: x != "_",
    )
    df = df.pipe(add_time).pipe(conc_units).pipe(poll_names)
    if polls is not None:
        df = df[df.poll.isin(polls)]
    if not in_domain(df, domain=domain).all():
        logger.warning("found obs outside the model domain")
        df = df[in_domain(df, domain=domain)]
    if (df.conc <= 0).any():
        logger.warning("found negative obs")
        df = df[df.conc > 0]
    return df["station lat lon alt time poll conc".split()]



"""
CAMS84 configuration
AeroCom PhaseIII optical properties experiment
"""

from collections import ChainMap

from pyaerocom.config import ALL_REGION_NAME

# Used below in some of the obsconfig entries to speed up when ran locally
OBS_ACCESS: dict = {}

# BASE FILTERS
ALTITUDE_FILTER = {"altitude": [0, 1000]}

GHOST_RURAL_FILTER = {
    "station_classification": ["background"],
    "area_classification": ["rural", "rural-near_city", "rural-regional", "rural-remote"],
}

EEA_NRT_RURAL_FILTER = {
    "station_classification": ["background"],
    "area_classification": ["rural", "rural-nearcity", "rural-regional", "rural-remote"],
}

# options station_classification: ['rural', 'urban', 'urban_bound']
# options area_classification: ['Agricultural', 'Commercial', 'Forested',
#               'Industrial', 'Residential', 'Undeveloped Rural', 'Unknown']
AIRNOW_RURAL_FILTER = {
    "station_classification": ["rural"]
    #'area_classification'     :   ['Agricultural', 'Forested', 'Undeveloped Rural']
}

# OBS SPECIFIC FILTERS (combination of the above and more)
GHOST_BASE_FILTER = {
    "set_flags_nan": True,
}

# How to read auxiliary model variables (that cannot be read but need to be
# computed). Entry under "fun" needs to be defined in ../eval_py/cube_read_methods.py
MODEL_AUX_VARS = {
    "vmro3": dict(vars_required=["mmro3"], fun="mmr_to_vmr"),
    "vmrno2": dict(vars_required=["mmrno2"], fun="mmr_to_vmr"),
}

# Setup for models used in analysis
MODELS = {
    "IFS-OSUITE": dict(model_id="ECMWF_OSUITE", model_read_aux=MODEL_AUX_VARS),
    "IFS-OSUITE-96h": dict(model_id="ECMWF_OSUITE_96H", model_read_aux=MODEL_AUX_VARS),
    "IFS-CTRL": dict(model_id="ECMWF_CNTRL", model_read_aux=MODEL_AUX_VARS),
    "IFS-CTRL-96h": dict(model_id="ECMWF_CNTRL_96H", model_read_aux=MODEL_AUX_VARS),
}

# Setup for available ground based observations (ungridded)

VAR_OUTLIER_RANGES = {
    "concpm10": [-1, 5000],  # ug m-3
    "concpm25": [-1, 5000],  # ug m-3
    "vmrno2": [-1, 5000],  # ppb
    "vmro3": [-1, 5000],  # ppb
}

AERONET_SITE_FILTER = dict(station_name="DRAGON*", negate="station_name")
OBS_GROUNDBASED = {
    "AeronetL1.5-d": dict(
        obs_id="AeronetSunV3Lev1.5.daily",
        obs_vars=["ang4487aer", "od550aer"],
        obs_vert_type="Column",
        obs_filters=ChainMap(ALTITUDE_FILTER, AERONET_SITE_FILTER),
        min_num_obs={"monthly": {"daily": 3}},
    ),
    "EEA-NRT-rural": dict(
        obs_id="EEAAQeRep.NRT",
        obs_data_dir=OBS_ACCESS.get("EEA_NRT_LOCAL"),
        obs_vars=["concpm10", "concpm25", "vmro3", "vmrno2"],
        obs_vert_type="Surface",
        obs_outlier_ranges=VAR_OUTLIER_RANGES,
        obs_filters=ChainMap(ALTITUDE_FILTER, EEA_NRT_RURAL_FILTER),
    ),
    "G-EEA-rural": dict(
        obs_id="GHOST.EEA.daily",
        obs_data_dir=OBS_ACCESS.get("GHOST_LOCAL_EEA_DAILY"),
        obs_vars=["concpm10", "concpm25", "vmro3", "vmrno2"],
        obs_vert_type="Surface",
        obs_outlier_ranges=VAR_OUTLIER_RANGES,
        obs_filters=ChainMap(ALTITUDE_FILTER, GHOST_BASE_FILTER, GHOST_RURAL_FILTER),
    ),
    "AirNow-rural": dict(
        obs_id="AirNow",
        obs_data_dir=None,
        obs_vars=["concpm10", "concpm25", "vmro3", "vmrno2"],
        obs_outlier_ranges=VAR_OUTLIER_RANGES,
        obs_filters=ChainMap(ALTITUDE_FILTER, AIRNOW_RURAL_FILTER),
        obs_vert_type="Surface",
    ),
    "AirNow": dict(
        obs_id="AirNow",
        obs_data_dir=OBS_ACCESS.get("AIRNOW_LOCAL"),
        obs_vars=["concpm10", "concpm25", "vmro3", "vmrno2"],
        obs_outlier_ranges=VAR_OUTLIER_RANGES,
        obs_filters={**ALTITUDE_FILTER},
        obs_vert_type="Surface",
    ),
    "MarcoPolo": dict(
        obs_id="MarcoPolo",
        obs_data_dir=OBS_ACCESS.get("MARCOPOLO_LOCAL"),
        obs_vars=["concpm10", "concpm25", "vmro3", "vmrno2"],
        obs_outlier_ranges=VAR_OUTLIER_RANGES,
        obs_vert_type="Surface",
    ),
}


# Setup for supported satellite evaluations
OBS_SAT: dict = {}

OBS_CFG = ChainMap(OBS_GROUNDBASED, OBS_SAT)


DEFAULT_RESAMPLE_CONSTRAINTS = dict(monthly=dict(daily=21), daily=dict(hourly=18))

CAMS84_CONFIG = dict(
    model_cfg=MODELS,
    obs_cfg=OBS_CFG,
    json_basedir="../../data",
    coldata_basedir="../../coldata",
    io_aux_file="../eval_py/gridded_io_aux.py",
    # if True, existing colocated data files will be deleted
    reanalyse_existing=True,
    only_json=False,
    add_model_maps=False,
    only_model_maps=False,
    clear_existing_json=False,
    # if True, the analysis will stop whenever an error occurs (else, errors that
    # occurred will be written into the logfiles)
    raise_exceptions=False,
    # Regional filter for analysis
    filter_name=f"{ALL_REGION_NAME}-wMOUNTAINS",
    # colocation frequency (no statistics in higher resolution can be computed)
    ts_type="daily",
    map_zoom="World",
    freqs=["daily", "monthly"],
    periods=["2018-2021", "2018", "2019", "2020", "2021"],
    main_freq="monthly",
    zeros_to_nan=False,
    min_num_obs=DEFAULT_RESAMPLE_CONSTRAINTS,
    colocate_time=False,
    obs_remove_outliers=True,
    model_remove_outliers=False,
    harmonise_units=True,
    regions_how="htap",
    annual_stats_constrained=False,
    proj_id="cams84",
    exp_id="eval",
    exp_name="Evaluation of CAMS forecast and reanalysis models",
    exp_descr=(
        "Both OSUITE and CNTRL are evaluated against multiple "
        "observation records including AOD from AERONET and "
        "PM, O3 and NO2 measurements."
    ),
    exp_pi="Jonas Gliß (jonasg@met.no)",
    public=True,
    # directory where colocated data files are supposed to be stored
    weighted_stats=True,
    var_order_menu=["od550aer", "ang4487aer", "concpm10", "concpm25", "vmrno2", "vmro3"],
)



from __future__ import annotations

from copy import deepcopy
from pathlib import Path

import pytest

from . import cfg_test_exp1, cfg_test_exp2, cfg_test_exp3, cfg_test_exp4, cfg_test_exp5

CFG_EXP: dict[str, dict] = dict(
    cfgexp1=cfg_test_exp1.CFG,
    cfgexp2=cfg_test_exp2.CFG,
    cfgexp3=cfg_test_exp3.CFG,
    cfgexp4=cfg_test_exp4.CFG,
    cfgexp5=cfg_test_exp5.CFG,
)


@pytest.fixture
def aeroval_out_path(tmp_path_factory) -> Path:
    """temporary path for aeroval outputs"""
    return tmp_path_factory.mktemp("aeroval")


@pytest.fixture(scope="session")
def aeroval_model_path(tmp_path_factory) -> Path:
    """temporary path for aeroval inputs"""
    return tmp_path_factory.mktemp("modeldata")


@pytest.fixture
def eval_config(cfg: str | None, aeroval_out_path: Path, aeroval_model_path: Path) -> dict:
    """aeroval configuration dispatcher"""
    if cfg is None:
        return {}
    if cfg not in CFG_EXP:
        raise ValueError(f"Unknown {cfg=}")
    config = deepcopy(CFG_EXP[cfg])
    config.update(
        json_basedir=f"{aeroval_out_path}/data",
        coldata_basedir=f"{aeroval_out_path}/coldata",
    )
    if cfg == "cfgexp3":
        config.update(
            model_cfg=cfg_test_exp3.fake_model_data(aeroval_model_path),
        )
    if cfg == "cfgexp5":
        config.update(
            model_cfg=cfg_test_exp5.fake_model_data(aeroval_model_path),
            obs_cfg=cfg_test_exp5.fake_obs_data(aeroval_model_path),
        )
    return config



import getpass
import logging
import os
from configparser import ConfigParser
from pathlib import Path

import numpy as np

from pyaerocom import obs_io
from pyaerocom._lowlevel_helpers import (
    check_dir_access,
    check_write_access,
    chk_make_subdir,
    list_to_shortstr,
)
from pyaerocom.data import resources
from pyaerocom.exceptions import DataIdError, DataSourceError
from pyaerocom.grid_io import GridIO
from pyaerocom.region_defs import ALL_REGION_NAME, HTAP_REGIONS, OLD_AEROCOM_REGIONS
from pyaerocom.varcollection import VarCollection
from pyaerocom.variable import Variable

logger = logging.getLogger(__name__)


class Config:
    """Class containing relevant paths for read and write routines

    A loaded instance of this class is created on import of pyaerocom and
    can be accessed via `pyaerocom.const`.

    TODO: provide more information
    """

    # NAMES
    # default names of the different obs networks
    # might get overwritten from paths.ini see func read_config

    #: ICP Forests
    ICPFORESTS_NAME = "ICPFORESTS"

    #: Aeronet Sun V2 access names
    AERONET_SUN_V2L15_AOD_DAILY_NAME = "AeronetSunV2Lev1.5.daily"
    AERONET_SUN_V2L15_AOD_ALL_POINTS_NAME = "AeronetSun_2.0_NRT"
    AERONET_SUN_V2L2_AOD_DAILY_NAME = "AeronetSunV2Lev2.daily"
    AERONET_SUN_V2L2_AOD_ALL_POINTS_NAME = "AeronetSunV2Lev2.AP"

    #: Aeronet SDA V2 access names
    AERONET_SUN_V2L2_SDA_DAILY_NAME = "AeronetSDAV2Lev2.daily"
    AERONET_SUN_V2L2_SDA_ALL_POINTS_NAME = "AeronetSDAV2Lev2.AP"

    # Aeronet V2 inversion products
    AERONET_INV_V2L15_DAILY_NAME = "AeronetInvV2Lev1.5.daily"
    AERONET_INV_V2L15_ALL_POINTS_NAME = "AeronetInvV2Lev1.5.AP"
    AERONET_INV_V2L2_DAILY_NAME = "AeronetInvV2Lev2.daily"
    AERONET_INV_V2L2_ALL_POINTS_NAME = "AeronetInvV2Lev2.AP"

    #: Aeronet Sun V3 access names
    AERONET_SUN_V3L15_AOD_DAILY_NAME = "AeronetSunV3Lev1.5.daily"
    AERONET_SUN_V3L15_AOD_ALL_POINTS_NAME = "AeronetSunV3Lev1.5.AP"
    AERONET_SUN_V3L2_AOD_DAILY_NAME = "AeronetSunV3Lev2.daily"
    AERONET_SUN_V3L2_AOD_ALL_POINTS_NAME = "AeronetSunV3Lev2.AP"

    #: Aeronet SDA V3 access names
    AERONET_SUN_V3L15_SDA_DAILY_NAME = "AeronetSDAV3Lev1.5.daily"
    AERONET_SUN_V3L15_SDA_ALL_POINTS_NAME = "AeronetSDAV3Lev1.5.AP"
    AERONET_SUN_V3L2_SDA_DAILY_NAME = "AeronetSDAV3Lev2.daily"
    AERONET_SUN_V3L2_SDA_ALL_POINTS_NAME = "AeronetSDAV3Lev2.AP"

    #: Aeronet V3 inversions
    AERONET_INV_V3L15_DAILY_NAME = "AeronetInvV3Lev1.5.daily"
    AERONET_INV_V3L2_DAILY_NAME = "AeronetInvV3Lev2.daily"

    # : CAMS2_83 name
    CAMS2_83_NRT_NAME = "CAMS2_83.NRT"

    #: EBAS name
    EBAS_MULTICOLUMN_NAME = "EBASMC"

    #: EEA nmea
    EEA_NAME = "EEAAQeRep"

    #: EEA.NRT name
    EEA_NRT_NAME = "EEAAQeRep.NRT"

    #: EEAV2 name
    EEA_V2_NAME = "EEAAQeRep.v2"

    #: Earlinet access name;
    EARLINET_NAME = "EARLINET"

    #: GAW TAD subset aas et al paper
    GAWTADSUBSETAASETAL_NAME = "GAWTADsubsetAasEtAl"

    #: DMS
    DMS_AMS_CVO_NAME = "DMS_AMS_CVO"

    #: CNEMC name (formally MEP)
    CNEMC_NAME = "CNEMC"

    #: ICOS name
    ICOS_NAME = "ICOS"

    # TROPOMI access names
    TROPOMI_XEMEP_R01x01_NAME = "TROPOMI_XEMEP_R01x01"

    # basename of paths.ini
    PATHS_INI_NAME = "paths.ini"

    #: boolean specifying wheter EBAS DB is copied to local cache for faster
    #: access, defaults to True
    EBAS_DB_LOCAL_CACHE = True

    #: Lowest possible year in data
    MIN_YEAR = 0
    #: Highest possible year in data
    MAX_YEAR = 20000

    #: standard names for coordinates
    STANDARD_COORD_NAMES = ["latitude", "longitude", "altitude"]
    #: Information specifying default vertical grid for post processing of
    #: profile data. The values are in units of m.
    DEFAULT_VERT_GRID_DEF = dict(lower=0, upper=15000, step=250)
    #: maximum allowed RH to be considered dry
    RH_MAX_PERCENT_DRY = 40

    DEFAULT_REG_FILTER = f"{ALL_REGION_NAME}-wMOUNTAINS"

    #: Time resample strategies for certain cominations, first level refers
    #: to TO, second to FROM and values are minimum number of observations
    OBS_MIN_NUM_RESAMPLE = dict(
        yearly=dict(monthly=3),
        monthly=dict(daily=7),
        daily=dict(hourly=6),
        hourly=dict(minutely=15),
    )

    #: This boolean can be used to enable / disable the former (i.e. use
    #: available wavelengths of variable in a certain range around variable
    #: wavelength).
    OBS_ALLOW_ALT_WAVELENGTHS = obs_io.OBS_ALLOW_ALT_WAVELENGTHS

    #: Wavelength tolerance for observations imports
    OBS_WAVELENGTH_TOL_NM = obs_io.OBS_WAVELENGTH_TOL_NM

    CLIM_START = 2005
    CLIM_STOP = 2015
    CLIM_FREQ = "daily"
    CLIM_RESAMPLE_HOW = "mean"  # median, ...
    # as a function of climatological frequency
    CLIM_MIN_COUNT = dict(
        daily=30,  # at least 30 daily measurements in each month over whole period
        monthly=5,  # analogue to daily ...
    )

    # names for the satellite data sets
    SENTINEL5P_NAME = "Sentinel5P"
    AEOLUS_NAME = "AeolusL2A"

    OLD_AEROCOM_REGIONS = OLD_AEROCOM_REGIONS

    URL_HTAP_MASKS = "https://pyaerocom.met.no/pyaerocom-suppl/htap_masks/"

    HTAP_REGIONS = HTAP_REGIONS

    RM_CACHE_OUTDATED = True

    #: Name of the file containing the revision string of an obs data network
    REVISION_FILE = "Revision.txt"

    #: timeout to check if one of the supported server locations can be
    #: accessed
    SERVER_CHECK_TIMEOUT = 1  # s

    def __init__(self, config_file=None, try_infer_environment=True):
        # Directories
        self._outputdir = None
        self._cache_basedir = None
        self._colocateddatadir = None
        self._logdir = None
        self._filtermaskdir = None
        self._local_tmp_dir = None
        self._downloaddatadir = None
        self._confirmed_access = []
        self._rejected_access = []

        # Options
        self._caching_active = True

        self._var_param = None
        self._coords = None

        # Custom variables
        self._custom_var_dict = None

        # Attributes that are used to store search directories
        self.OBSLOCS_UNGRIDDED = {}
        self.OBS_UNGRIDDED_POST = {}
        self.SUPPLDIRS = {}
        self._search_dirs = []

        self.WRITE_FILEIO_ERR_LOG = True

        self.last_config_file = None
        self._ebas_flag_info = None

        self._outhomename = "MyPyaerocom"

        with resources.path("pyaerocom.data", "paths.ini") as path:
            self._config_ini_lustre = str(path)
        with resources.path("pyaerocom.data", "variables.ini") as path:
            self._var_info_file = str(path)
        with resources.path("pyaerocom.data", "coords.ini") as path:
            self._coords_info_file = str(path)

        self._user = getpass.getuser()
        self.my_pyaerocom_dir = os.path.join(f"{os.path.expanduser('~')}", self._outhomename)

        self.DO_NOT_CACHE_FILE = None

        self.ERA5_SURFTEMP_FILENAME = "era5.msl.t2m.201001-201012.nc"

        self._LUSTRE_CHECK_PATH = "/"

        #: Settings for reading and writing of gridded data
        self.GRID_IO = GridIO()

        if config_file is not None:
            if not os.path.exists(config_file):
                raise FileNotFoundError(f"input config file does not exist {config_file}")
            elif not config_file.endswith("ini"):
                raise ValueError("Need path to an ini file for input config_file")

            basedir, config_file = os.path.split(config_file)
        elif try_infer_environment:
            try:
                config_file = self.infer_config()
            except FileNotFoundError:
                pass

        if config_file is not None:
            try:
                self.read_config(
                    config_file,
                )
            except Exception as e:
                logger.warning(f"Failed to read config. Error: {repr(e)}")
        # create MyPyaerocom directory
        chk_make_subdir(self.HOMEDIR, self._outhomename)

    def _check_access(self, loc, timeout=None):
        """Uses multiprocessing approach to check if location can be accessed

        Parameters
        ----------
        loc : str
            path that is supposed to be checked

        Returns
        -------
        bool
            True, if location is accessible, else False
        """
        if loc is None:
            return False
        loc = str(Path(loc))  # make sure the path is set correctly
        if loc in self._confirmed_access:
            return True
        elif loc in self._rejected_access:
            return False

        logger.info(f"Checking access to: {loc}")
        if check_dir_access(loc):
            self._confirmed_access.append(loc)
            return True
        self._rejected_access.append(loc)
        return False

    def _basedirs_search_db(self):
        return [self.ROOTDIR, self.HOMEDIR]

    def _infer_config_from_basedir(self, basedir):
        basedir = os.path.normpath(basedir)
        for env_id, chk_sub in self._check_subdirs_cfg.items():
            chkdir = os.path.join(basedir, chk_sub)
            if self._check_access(chkdir):
                return (self._config_files[env_id], env_id)

        raise FileNotFoundError(
            f"Could not infer environment configuration for input directory: {basedir}"
        )

    def infer_config(self):
        """
        check if ~/MyPyaerocom/paths.ini exists.
        if not, use the default paths.ini
        """

        self._paths_ini = os.path.join(self.my_pyaerocom_dir, self.PATHS_INI_NAME)
        if os.path.exists(self._paths_ini):
            logger.info(f"using user specific config file: {self._paths_ini}")
        else:
            with resources.path("pyaerocom.data", self.PATHS_INI_NAME) as path:
                self._paths_ini = str(path)
                logger.info(f"using default config file: {self._paths_ini}")

        return self._paths_ini

    def register_custom_variables(
        self, vars: dict[str, Variable] | dict[str, dict[str, str]]
    ) -> None:
        var_dict = {}
        for key, item in vars.items():
            if isinstance(item, Variable):
                var_dict[key] = item
            elif isinstance(item, dict):
                if "var_name" in item and "units" in item:
                    var_dict[key] = Variable(**item)
                else:
                    raise ValueError(
                        f"Dict item {item} must at least have the keys 'var_name' and 'units'"
                    )
            else:
                raise ValueError(f"Item {item} must be either dict or Variable")
        self._custom_var_dict = var_dict.copy()
        self._var_param = None

    @property
    def has_access_users_database(self):
        chk_dir = self._check_subdirs_cfg["users-db"]
        chk_paths = [
            os.path.join("/metno/aerocom_users_database/", chk_dir),
            os.path.join(self.HOMEDIR, "/aerocom_users_database/", chk_dir),
        ]
        for p in chk_paths:
            if self._check_access(p):
                return True
        return False

    @property
    def has_access_lustre(self):
        """Boolean specifying whether MetNO AeroCom server is accessible"""
        for path in self._search_dirs:
            if self._LUSTRE_CHECK_PATH in path and self._check_access(path):
                return True
        return False

    @property
    def ALL_DATABASE_IDS(self):
        """ID's of available database configurations"""
        return list(self._config_files)

    @property
    def ROOTDIR(self):
        """Local root directory"""
        return os.path.abspath(os.sep)

    @property
    def HOMEDIR(self):
        """Home directory of user"""
        return os.path.expanduser("~") + "/"

    @property
    def OUTPUTDIR(self):
        """Default output directory"""
        self._outputdir = chk_make_subdir(self.HOMEDIR, self._outhomename)
        return self._outputdir

    @property
    def DATA_SEARCH_DIRS(self):
        """
        Directories which pyaerocom will consider for data access

        Note
        ----
        This corresponds to directories considered for searching gridded
        data (e.g. models and level 3 satellite products).  Please
        see :attr:`OBSLOCS_UNGRIDDED` for available data directories
        for reading of ungridded data.

        Returns
        -------
        list
            list of directories

        """
        return self._search_dirs

    @property
    def FILTERMASKKDIR(self):
        if not check_write_access(self._filtermaskdir):
            outdir = self.OUTPUTDIR
            self._filtermaskdir = chk_make_subdir(outdir, "filtermasks")
        return self._filtermaskdir

    @property
    def COLOCATEDDATADIR(self):
        """Directory for accessing and saving colocated data objects"""
        if not check_write_access(self._colocateddatadir):
            outdir = self.OUTPUTDIR
            self._colocateddatadir = chk_make_subdir(outdir, "colocated_data")
        return self._colocateddatadir

    @property
    def LOCAL_TMP_DIR(self):
        """Local TEMP directory"""
        if self._local_tmp_dir is None:
            self._local_tmp_dir = f"{self.OUTPUTDIR}/tmp"
        if not self._check_access(self._local_tmp_dir):
            os.makedirs(self._local_tmp_dir, exist_ok=True)
        return self._local_tmp_dir

    @LOCAL_TMP_DIR.setter
    def LOCAL_TMP_DIR(self, val):
        self._local_tmp_dir = val

    @property
    def DOWNLOAD_DATADIR(self):
        """Directory where data is downloaded into"""
        if self._downloaddatadir is None:
            self._downloaddatadir = chk_make_subdir(self.OUTPUTDIR, "data")
        return self._downloaddatadir

    @DOWNLOAD_DATADIR.setter
    def DOWNLOAD_DATADIR(self, val):
        if not isinstance(val, str):
            raise ValueError("Please provide str")
        elif not os.path.exists(val):
            try:
                os.mkdir(val)
            except Exception:
                raise OSError(f"Input directory {val} does not exist and can also not be created")
        self._downloaddatadir = val

    @property
    def user(self):
        """User ID"""
        return getpass.getuser()

    @property
    def cache_basedir(self):
        """Base directory for caching

        The actual files are cached in user subdirectory, cf :attr:`CACHEDIR`
        """
        cd = self._cache_basedir
        if not check_write_access(cd):
            outdir = self.OUTPUTDIR
            cd = chk_make_subdir(outdir, "_cache")
            self._cache_basedir = cd
        return cd

    @cache_basedir.setter
    def cache_basedir(self, val):
        if not check_write_access(val):
            raise ValueError(val)
        self._cache_basedir = os.path.abspath(val)

    @property
    def CACHEDIR(self):
        """Cache directory for UngriddedData objects"""
        try:
            return chk_make_subdir(self.cache_basedir, self.user)
        except Exception as e:
            logger.warning(f"Failed to access CACHEDIR: {repr(e)}\nDeactivating caching")
            self._caching_active = False

    @CACHEDIR.setter
    def CACHEDIR(self, val):
        """Cache directory"""
        if not check_write_access(val):
            raise ValueError(
                f"Cannot set cache directory. "
                f"Input directory {val} does not exist or write permission is not granted"
            )
        spl = os.path.split(val)
        if spl[-1] == self.user:
            val = spl[0]
        self._cache_basedir = val

    @property
    def CACHING(self):
        """Activate writing of and reading from cache files"""
        return self._caching_active

    @CACHING.setter
    def CACHING(self, val):
        self._caching_active = bool(val)

    @property
    def VAR_PARAM(self):
        """Deprecated name, please use :attr:`VARS` instead"""
        logger.warning("Deprecated (but still functional) name VAR_PARAM. Please use VARS")
        return self.VARS

    @property
    def VARS(self):
        """Instance of class VarCollection (for default variable information)"""
        if self._var_param is None:  # has not been accessed before
            self._var_param = VarCollection(self._var_info_file)

            if self._custom_var_dict is not None:
                for var in self._custom_var_dict:
                    self._var_param.add_var(self._custom_var_dict[var])
        return self._var_param

    @property
    def COORDINFO(self):
        """Instance of :class:`VarCollection` containing coordinate info"""
        if self._coords is None:
            self._coords = VarCollection(self._coords_info_file)
        return self._coords

    @property
    def LOGFILESDIR(self):
        """Directory where logfiles are stored"""
        if self._logdir is None:
            self._logdir = chk_make_subdir(self.OUTPUTDIR, "_log")
        return self._logdir

    @property
    def ETOPO1_AVAILABLE(self):
        """
        Boolean specifying if access to ETOPO1 dataset is provided

        Returns
        -------
        bool
        """
        if "etopo1" in self.SUPPLDIRS and os.path.exists(self.SUPPLDIRS["etopo1"]):
            return True
        return False

    @property
    def EBAS_FLAGS_FILE(self):
        """Location of CSV file specifying meaning of EBAS flags"""
        with resources.path("pyaerocom.data", "ebas_flags.csv") as path:
            return str(path)

    @property
    def OBS_IDS_UNGRIDDED(self):
        """List of all data IDs of supported ungridded observations"""
        ids = list(self.OBSLOCS_UNGRIDDED)
        ids.extend(self.OBS_UNGRIDDED_POST)
        return ids

    @property
    def ERA5_SURFTEMP_FILE(self):
        if "era5" in self.SUPPLDIRS:
            sdir = self.SUPPLDIRS["era5"]
            if os.path.exists(sdir) and self.ERA5_SURFTEMP_FILENAME in os.listdir(sdir):
                return os.path.join(sdir, self.ERA5_SURFTEMP_FILENAME)
        raise FileNotFoundError(
            "ERA Interim surface temperature data cannot be accessed (check lustre connection)"
        )

    def make_default_vert_grid(self):
        """Makes default vertical grid for resampling of profile data"""
        step = self.DEFAULT_VERT_GRID_DEF["step"]
        offs = int(step / 2)
        return np.arange(
            self.DEFAULT_VERT_GRID_DEF["lower"] + offs,
            self.DEFAULT_VERT_GRID_DEF["upper"] - offs,
            step,
        )

    def add_data_search_dir(self, *dirs):
        """Add data search directories for database browsing"""
        for loc in dirs:
            if not self._check_access(loc):
                raise FileNotFoundError(f"Input location {loc} could not be accessed")
            self._search_dirs.append(loc)

    def add_ungridded_obs(self, obs_id, data_dir, reader=None, check_read=False):
        """Add a network to the data search structure

        Parameters
        ----------
        obs_id : str
            name of network. E.g. MY_OBS or EBASMC
        data_dir : str
            directory where data files are stored
        reader : pyaerocom.io.ReadUngriddedBase, optional
            reading class used to import these data. If `obs_id` is known
            (e.g. EBASMC) this is not needed.

        Raises
        ------
        AttributeError
            if the network name is already reserved in :attr:`OBSLOCS_UNGRIDDED`
        ValueError
            if the data directory does not exist
        """
        if obs_id in self.OBS_IDS_UNGRIDDED:
            raise DataIdError(
                f"Network with ID {obs_id} is already registered at "
                f"{self.OBSLOCS_UNGRIDDED[obs_id]}"
            )
        elif not self._check_access(data_dir):
            raise ValueError("Input data directory cannot be accessed")
        if reader is None:
            from pyaerocom.io.utils import get_ungridded_reader

            reader = get_ungridded_reader(obs_id)

        if obs_id not in reader.SUPPORTED_DATASETS:
            reader.SUPPORTED_DATASETS.append(obs_id)
        self.OBSLOCS_UNGRIDDED[obs_id] = data_dir
        if check_read:
            self._check_obsreader(obs_id, data_dir, reader)

    def add_ungridded_post_dataset(
        self,
        obs_id,
        obs_vars,
        obs_aux_requires,
        obs_merge_how,
        obs_aux_funs=None,
        obs_aux_units=None,
        **kwargs,
    ):
        """
        Register new ungridded dataset

        Other than :func:`add_ungridded_obs`, this method adds required logic
        for a "virtual" ungridded observation datasets, that is, a dataset that
        can only be computed from other ungridded datasets but not read from
        disk.

        If all input parameters are okay, the new dataset will be registered
        in :attr:`OBS_UNGRIDDED_POST` and will then be accessible for import
        in ungridded reading factory class :class:`pyaerocom.io.ReadUngridded`.

        Parameters
        ----------
        obs_id : str
            Name of new dataset.
        obs_vars : str or list
            variables supported by this dataset.
        obs_aux_requires : dict
            dicionary specifying required datasets and variables for each
            variable supported by the auxiliary dataset.
        obs_merge_how : str or dict
            info on how to derive each of the supported coordinates (e.g. eval,
            combine). For valid input args see
            :mod:`pyaerocom.combine_vardata_ungridded`. If value is string,
            then the same method is used for all variables.
        obs_aux_funs : dict, optional
            dictionary specifying computation methods for auxiliary variables
            that are supposed to be retrieved via `obs_merge_how='eval'`.
            Keys are variable names, values are respective computation methods
            (which need to be strings as they will be evaluated via
             :func:`pandas.DataFrame.eval` in
             :mod:`pyaerocom.combine_vardata_ungridded`). This input is
            optional, but mandatory if any of the `obs_vars` is
            supposed to be retrieved via `merge_how='eval'`.
        obs_aux_units : dict, optional
            output units of auxiliary variables (only needed for varibales
            that are derived via `merge_how='eval'`)
        **kwargs
            additional keyword arguments (unused, but serves the purpose to
            allow for parsing info from dictionaries and classes that
            contain additional attributes than the ones needed here).

        Raises
        ------
        ValueError
            if input obs_id is already reserved


        Returns
        -------
        None.

        """
        if obs_id in self.OBS_IDS_UNGRIDDED:
            raise ValueError(f"Network with ID {obs_id} is already registered...")
        elif obs_aux_units is None:
            obs_aux_units = {}
        # this class will do the required sanity checking and will only
        # initialise if everything is okay
        addinfo = obs_io.AuxInfoUngridded(
            data_id=obs_id,
            vars_supported=obs_vars,
            aux_requires=obs_aux_requires,
            aux_merge_how=obs_merge_how,
            aux_funs=obs_aux_funs,
            aux_units=obs_aux_units,
        )

        self.OBS_UNGRIDDED_POST[obs_id] = addinfo.to_dict()

    def _check_obsreader(self, obs_id, data_dir, reader):
        """
        Check if files can be accessed when registering new dataset

        Parameters
        ----------
        obs_id : str
            name of obsnetwork
        data_dir : str
            directory containing data files
        reader : ReadUngriddedBase
            reading interface
        """
        check = reader(obs_id)
        path = check.data_dir
        assert path == data_dir
        try:
            check.get_file_list()
        except DataSourceError:
            if "renamed" not in os.listdir(data_dir):
                raise
            logger.warning(
                f"Failed to register {obs_id} at {data_dir} using ungridded "
                f"reader {reader} but input dir has a renamed subdirectory, "
                f"trying to find valid data files in there instead"
            )
            chk_dir = os.path.join(data_dir, "renamed")
            self.OBSLOCS_UNGRIDDED.pop(obs_id)
            self.add_ungridded_obs(obs_id, chk_dir, reader, check_read=True)

    @property
    def ebas_flag_info(self):
        """Information about EBAS flags

        Note
        ----
        Is loaded upon request -> cf.
        :attr:`pyaerocom.io.ebas_nasa_ames.EbasFlagCol.FLAG_INFO`

        Dictionary containing 3 dictionaries (keys: ```valid, values, info```)
        that contain information about validity of each flag (```valid```),
        their actual values (```values```, e.g. V, M, I)
        """
        if self._ebas_flag_info is None:
            from pyaerocom.io.helpers import read_ebas_flags_file

            self._ebas_flag_info = read_ebas_flags_file(self.EBAS_FLAGS_FILE)
        return self._ebas_flag_info

    def reload(self, keep_basedirs=True):
        """Reload config file (for details see :func:`read_config`)"""
        self.read_config(self.last_config_file, keep_basedirs)

    def read_config(
        self,
        config_file,
        basedir=None,
        init_obslocs_ungridded=False,
        init_data_search_dirs=False,
    ):
        """
        Import paths from one of the config ini files

        Parameters
        ----------
        config_file : str
            file location of config ini file
        basedir : str, optional
            Base directory to be used for relative model and obs dirs specified
            via BASEDIR in config file. If None, then the BASEDIR value in the
            config file is used. The default is None.
        init_obslocs_ungridded : bool, optional
            If True, :attr:`OBSLOCS_UNGRIDDED` will be re-instantiated (i.e.
            all currently set obs locations will be deleted).
            The default is False.
        init_data_search_dirs : bool, optional
            If True, :attr:`DATA_SEARCH_DIRS` will be re-instantiated (i.e.
            all currently set data search directories will be deleted).
            The default is False.

        Raises
        ------
        FileNotFoundError
            If input config file is not a file or does not exist.

        Returns
        -------
        None.

        """

        if not os.path.isfile(config_file):
            raise FileNotFoundError(
                f"Configuration file paths.ini at {config_file} does not exist "
                f"or is not a file"
            )

        if init_obslocs_ungridded:
            self.OBSLOCS_UNGRIDDED = {}
        if init_data_search_dirs:
            self._search_dirs = []

        cr = ConfigParser()
        cr.optionxform = str
        cr.read(config_file)
        # init base directories for Model data
        if cr.has_section("modelfolders"):
            self._add_searchdirs(cr, basedir)

        if cr.has_section("obsfolders"):
            self._add_obsconfig(cr, basedir)

        if cr.has_section("outputfolders"):
            self._init_output_folders_from_cfg(cr)

        if cr.has_section("supplfolders"):
            if basedir is None and "BASEDIR" in cr["supplfolders"]:
                basedir = cr["supplfolders"]["BASEDIR"]

            for name, path in cr["supplfolders"].items():
                if "${BASEDIR}" in path:
                    path = path.replace("${BASEDIR}", basedir)
                self.SUPPLDIRS[name] = path

        cr.clear()
        self.GRID_IO.load_aerocom_default()
        self.last_config_file = config_file

    def _resolve_basedir(self, locs, chk_dirs):
        repl = "${BASEDIR}"
        for loc in locs:
            if repl in loc:
                for chk_dir in chk_dirs:
                    chk = Path(loc.replace(repl, chk_dir))
                    if self._check_access(chk):
                        return chk_dir
        raise FileNotFoundError("Could not confirm any directory...")

    def _add_searchdirs(self, cr, basedir=None):
        chk_dirs = []
        if basedir is not None and self._check_access(basedir):
            chk_dirs.append(basedir)

        mcfg = cr["modelfolders"]

        # check and update model base directory if applicable
        if "BASEDIR" in mcfg:
            _dir = mcfg["BASEDIR"]
            if "${HOME}" in _dir:
                _dir = _dir.replace("${HOME}", os.path.expanduser("~"))
            elif "${USER}" in _dir:
                _dir = _dir.replace("${USER}", self._user)
            if _dir not in chk_dirs and self._check_access(_dir):
                chk_dirs.append(_dir)
        if len(chk_dirs) == 0:
            return False

        # get all locations defined in config file as list
        locs = mcfg["dir"].replace("\n", "").split(",")

        # find first location that contains BASEDIR to determine
        try:
            basedir = str(self._resolve_basedir(locs, chk_dirs))
        except FileNotFoundError:
            basedir = None
        repl_str = "${BASEDIR}"
        for loc in locs:
            if repl_str in loc:
                if basedir is None:
                    continue
                loc = loc.replace(repl_str, basedir)

            if loc not in self._search_dirs:
                self._search_dirs.append(loc)
        return True

    def _add_obsconfig(self, cr, basedir=None):
        chk_dirs = []
        if basedir is not None and self._check_access(basedir):
            chk_dirs.append(basedir)

        cfg = cr["obsfolders"]

        # check and update model base directory if applicable
        if "BASEDIR" in cfg:
            _dir = cfg["BASEDIR"]
            if "${HOME}" in _dir:
                _dir = _dir.replace("${HOME}", os.path.expanduser("~"))
            if "${USER}" in _dir:
                _dir = _dir.replace("${USER}", self._user)
            if _dir not in chk_dirs and self._check_access(_dir):
                chk_dirs.append(_dir)
        if len(chk_dirs) == 0:
            return False

        names_cfg = self._add_obsnames_config(cr)

        candidates = {}
        dirconfirmed = None
        repl = "${BASEDIR}"
        if cr.has_section("obsfolders"):
            for obsname, path in cr["obsfolders"].items():
                if obsname.lower() == "basedir":
                    continue
                name_str = f"{obsname.upper()}_NAME"
                if name_str in names_cfg:
                    ID = self.__dict__[name_str]
                else:
                    ID = self._add_obsname(obsname)
                candidates[ID] = path
                # candidate for checking access
                if dirconfirmed is None and repl in path:
                    for chk_dir in chk_dirs:
                        chk = Path(path.replace(repl, chk_dir))
                        if self._check_access(chk):
                            dirconfirmed = str(chk_dir)
        homedir = os.path.expanduser("~")
        for name, loc in candidates.items():
            if repl in loc:
                if dirconfirmed is None:
                    continue
                loc = loc.replace(repl, dirconfirmed)
            if "${HOME}" in loc:
                loc = loc.replace("${HOME}", homedir)

            self.OBSLOCS_UNGRIDDED[name] = loc

    def _init_output_folders_from_cfg(self, cr):
        cfg = cr["outputfolders"]
        if "CACHEDIR" in cfg and not self._check_access(self._cache_basedir):
            self._cache_basedir = cfg["CACHEDIR"]

        if "OUTPUTDIR" in cfg and not self._check_access(self._outputdir):
            self._outputdir = cfg["OUTPUTDIR"]

        if "COLOCATEDDATADIR" in cfg and not self._check_access(self._colocateddatadir):
            self._colocateddatadir = cfg["COLOCATEDDATADIR"]

        if "LOCALTMPDIR" in cfg:
            _dir = cfg["LOCALTMPDIR"]
            # expand ${HOME}
            if "${HOME}" in _dir:
                _dir = _dir.replace("${HOME}", os.path.expanduser("~"))
            if "${USER}" in _dir:
                _dir = _dir.replace("${USER}", getpass.getuser())

            self._local_tmp_dir = _dir

    def _add_obsname(self, name):
        name_str = f"{name.upper()}_NAME"
        self[name_str] = name
        return name_str

    def _add_obsnames_config(self, cr):
        names_cfg = []
        if cr.has_section("obsnames"):
            for obsname, ID in cr["obsnames"].items():
                name_str = f"{obsname.upper()}_NAME"
                self[name_str] = ID
                names_cfg.append(name_str)
        return names_cfg

    def short_str(self):
        """Deprecated method"""
        return self.__str__()

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self):
        head = f"Pyaerocom {type(self).__name__}"
        s = f"\n{head}\n{len(head) * '-'}\n"
        for k, v in self.__dict__.items():
            if k.startswith("_"):
                pass
            if k == "VARS":
                s += f"\n{k}\n{list_to_shortstr(v.all_vars)}"
            elif isinstance(v, dict):
                s += f"\n{k} (dict)"
            elif isinstance(v, list):
                s += f"\n{k} (list)"
                s += list_to_shortstr(v)
            else:
                s += f"\n{k}: {v}"
        return s



from __future__ import annotations

import numpy as np
import pandas as pd
import pytest

from pyaerocom import StationData


def station_data1() -> StationData:
    stat = StationData()

    d = dict(
        dataset_name="test",
        PI="jonas gliss",
        instrument_name="test instr",
        station_id=42,
        station_name="test station",
        ts_type="monthly",
        revision_date="20190513",
        data_level=2,
        country="norway",
        random_key1="bla",
        random_key2="blub",
        random_key3=["blub", "blablub"],
        random_key4=["blub", "blub"],
        filename="bla.csv",
        data_version=2,
    )

    stat.update(d)

    START = "2000"
    NUM_YEARS = 5
    NUM = NUM_YEARS * 12
    stat.dtime = np.datetime64(START) + np.arange(NUM).astype("timedelta64[M]")

    stat.ec550aer = np.random.random_sample(NUM) - 0.5
    stat.od550aer = np.ones(NUM)
    stat.latitude = list(np.ones(NUM) * 33)
    stat.longitude = np.ones(NUM) * 15
    stat.altitude = (np.ones(NUM) * 300).astype(int)

    stat.var_info["ec550aer"] = {"units": "m-1"}
    stat.var_info["od550aer"] = {"units": "1"}

    return stat


def station_data2() -> StationData:
    """Create an example synthetic instance of StationData class"""
    stat = StationData()
    d = dict(
        latitude=33.01,
        longitude=15,
        altitude=300,
        dataset_name="test (alt)",
        PI="Jonas Gliss",
        instrument_name="test instr",
        station_id=42,
        station_name="test station",
        ts_type="daily",
        revision_date="20190513",
        data_level=3,
        country="norway",
        data_version=2,
    )

    stat.update(d)

    START = "2007"
    NUM_DAYS = 277

    stat.dtime = np.datetime64(START) + np.arange(NUM_DAYS).astype("timedelta64[D]")

    stat.ec550aer = np.random.random_sample(NUM_DAYS) - 0.5
    stat.od550aer = np.ones(NUM_DAYS)
    stat.conco3 = np.arange(NUM_DAYS)

    stat.var_info["ec550aer"] = {"units": "Mm-1"}
    stat.var_info["od550aer"] = {"units": "1"}
    stat.var_info["conco3"] = {"units": "ug m-3"}

    return stat


class FakeStationDataAccess:
    """Factory for loading and accessing of data objects"""

    _LOADERS = dict(station_data1=station_data1, station_data2=station_data2)

    def __getitem__(self, key):
        if key in self.__dict__:  # item is loaded
            return self.__dict__[key]
        data = self._LOADERS[key]()
        self.__dict__[key] = data
        return data


FAKE_STATION_DATA = FakeStationDataAccess()


def create_fake_station_data(addvars, varinfo, varvals, start, stop, freq, meta) -> StationData:
    if isinstance(addvars, str):
        addvars = [addvars]
    stat = StationData()
    stat.update(**meta)
    dtime = pd.date_range(start, stop, freq=freq).values
    stat["dtime"] = dtime
    for var in addvars:
        if var in varinfo:
            stat.var_info[var] = varinfo[var]
        if isinstance(varvals, dict):
            val = varvals[var]
        else:
            val = varvals
        stat[var] = np.ones(len(dtime)) * val
    return stat


def create_fake_stationdata_list() -> list[StationData]:
    stat1 = create_fake_station_data(
        "concpm10",
        {"concpm10": {"units": "ug m-3"}},
        10,
        "2010-01-01",
        "2010-12-31",
        "d",
        {
            "awesomeness": 10,
            "data_revision": 20120101,
            "ts_type": "daily",
            "latitude": 42.001,
            "longitude": 20,
            "altitude": 0.1,
            "station_name": "FakeSite",
        },
    )

    stat2 = create_fake_station_data(  # overlaps with first one
        "concpm10",
        {"concpm10": {"units": "ug m-3"}},
        20,
        "2010-06-01",
        "2011-12-31",
        "d",
        {
            "awesomeness": 12,
            "data_revision": 20110101,
            "ts_type": "daily",
            "latitude": 42.001,
            "longitude": 20,
            "altitude": 0.1,
            "station_name": "FakeSite",
        },
    )

    stat3 = create_fake_station_data(  # monthly, but missing ts_type and wrong unit
        "concpm10",
        {"concpm10": {"units": "mole mole-1"}},
        20,
        "2014-01-01",
        "2015-12-31",
        "3MS",
        {
            "awesomeness": 2,
            "data_revision": 20140101,
            "latitude": 42.001,
            "longitude": 20,
            "altitude": 0.1,
            "station_name": "FakeSite",
        },
    )

    stat4 = create_fake_station_data(  # invalid ts_type
        "concpm10",
        {"concpm10": {"units": "ug m-3"}},
        20,
        "1850",
        "2020",
        "1000d",
        {
            "awesomeness": 15,
            "data_revision": 20130101,
            "ts_type": "1000daily",
            "latitude": 42.001,
            "longitude": 20,
            "altitude": 0.1,
            "station_name": "FakeSite",
        },
    )

    stat5 = create_fake_station_data(  # new variable and monthly
        "od550aer",
        {"od550aer": {"units": "1"}},
        1,
        "2005",
        "2012",
        "MS",
        {
            "awesomeness": 42,
            "data_revision": 20200101,
            "ts_type": "monthly",
            "latitude": 22.001,
            "longitude": 10,
            "altitude": 99,
            "station_name": "FakeSite",
        },
    )

    stat6 = create_fake_station_data(
        "od550aer",
        {"od550aer": {"units": "1"}},
        0.1,
        "2008",
        "2009",
        "60d",
        {
            "awesomeness": 46,
            "data_revision": 20200101,
            "ts_type": "60daily",
            "latitude": 22.001,
            "longitude": 10,
            "altitude": 100,
            "station_name": "FakeSite2",
        },
    )

    stat_werr = create_fake_station_data(
        "od550aer",
        {"od550aer": {"units": "1"}},
        0.2,
        "2010",
        "2016",
        "10d",
        {
            "awesomeness": 30,
            "data_revision": 20200101,
            "ts_type": "10daily",
            "latitude": 22.001,
            "longitude": 10,
            "altitude": 100,
            "station_name": "FakeSite2",
        },
    )
    stat_werr.data_err["od550aer"] = np.ones(len(stat_werr.dtime)) * 9999

    return [stat1, stat2, stat3, stat4, stat5, stat6, stat_werr]


@pytest.fixture(scope="session")
def statlist():
    """fake stations data"""
    data = {}
    stats = create_fake_stationdata_list()
    data["all"] = stats
    data["od550aer"] = [stat.copy() for stat in stats if stat.has_var("od550aer")]
    pm10sites = [stat.copy() for stat in stats if stat.has_var("concpm10")]
    data["concpm10_X"] = pm10sites
    data["concpm10_X2"] = [stat.copy() for stat in pm10sites[:3]]
    data["concpm10"] = [stat.copy() for stat in pm10sites[:2]]
    return data



"""
Mathematical low level utility methods of pyaerocom
"""

import numpy as np
from scipy.stats import pearsonr

from pyaerocom._warnings import ignore_warnings


def in_range(x, low, high) -> bool:
    return low <= x <= high


def is_strictly_monotonic(iter1d) -> bool:
    """
    Check if 1D iterable is strictly monotonic

    Parameters
    ----------
    iter1d
        1D iterable object to be tested

    Returns
    -------
    bool

    """
    return True if np.all(np.diff(iter1d) > 0) else False


def make_binlist(vmin: float, vmax: float, num: int = 8) -> list[float]:
    return np.linspace(vmin, vmax, num + 1).tolist()


def weighted_sum(data, weights):
    """Compute weighted sum using numpy dot product

    Parameters
    ----------
    data : ndarray
        data array that is supposed to be summed up
    weights : ndarray
        array containing weights for each point in `data`

    Returns
    -------
    float
        weighted sum of values in input array
    """
    return np.dot(data, weights)


def sum(data, weights=None):
    """Summing operation with option to perform weighted sum

    Parameters
    ----------
    data : ndarray
        data array that is supposed to be summed up
    weights : ndarray, optional
        array containing weights for each point in `data`

    Returns
    -------
    float or int
        sum of values in input array
    """
    if weights is None:
        return np.sum(data)
    return weighted_sum(data, weights)


def weighted_mean(data, weights):
    """Compute weighted mean

    Parameters
    ----------
    data : ndarray
        data array that is supposed to be averaged
    weights : ndarray
        array containing weights for each point in `data`

    Returns
    -------
    float or int
        weighted mean of data array
    """
    return np.sum(data * weights) / np.sum(weights)


def weighted_cov(ref_data, data, weights):
    """Compute weighted covariance

    Parameters
    ----------
    data_ref : ndarray
        x data
    data : ndarray
        y data
    weights : ndarray
        array containing weights for each point in `data`

    Returns
    -------
    float
        covariance
    """
    avgx = weighted_mean(ref_data, weights)
    avgy = weighted_mean(data, weights)
    return np.sum(weights * (ref_data - avgx) * (data - avgy)) / np.sum(weights)


@ignore_warnings(RuntimeWarning, "invalid value encountered in scalar divide")
def weighted_corr(ref_data, data, weights):
    """Compute weighted correlation

    Parameters
    ----------
    data_ref : ndarray
        x data
    data : ndarray
        y data
    weights : ndarray
        array containing weights for each point in `data`

    Returns
    -------
    float
       weighted correlation coefficient
    """
    wcovxy = weighted_cov(ref_data, data, weights)

    wcovxx = weighted_cov(ref_data, ref_data, weights)
    wcovyy = weighted_cov(data, data, weights)
    wsigmaxy = np.sqrt(wcovxx * wcovyy)
    return wcovxy / wsigmaxy


@ignore_warnings(
    RuntimeWarning, "invalid value encountered in double_scalars", "An input array is constant"
)
def corr(ref_data, data, weights=None):
    """Compute correlation coefficient

    Parameters
    ----------
    data_ref : ndarray
        x data
    data : ndarray
        y data
    weights : ndarray, optional
        array containing weights for each point in `data`

    Returns
    -------
    float
       correlation coefficient
    """
    if weights is None:
        return pearsonr(ref_data, data)[0]
    return weighted_corr(ref_data, data, weights)


def closest_index(num_array, value):
    """Returns index in number array that is closest to input value"""
    return np.argmin(np.abs(np.asarray(num_array) - value))


def numbers_in_str(input_string):
    """This method finds all numbers in a string

    Note
    ----
    - Beta version, please use with care
    - Detects only integer numbers, dots are ignored

    Parameters
    ----------
    input_string : str
        string containing numbers

    Returns
    -------
    list
        list of strings specifying all numbers detected in string

    Example
    -------
    >>> numbers_in_str('Bla42Blub100')
    [42, 100]
    """
    numbers = []
    IN_NUM = False
    c_num = None
    for char in input_string:
        try:
            int(char)
            if not IN_NUM:
                IN_NUM = True
                c_num = char
            elif IN_NUM:
                c_num += char
        except Exception:
            if IN_NUM:
                numbers.append(c_num)
            IN_NUM = False
    if IN_NUM:
        numbers.append(c_num)
    return numbers


def exponent(num):
    """Get exponent of input number

    Parameters
    ----------
    num : :obj:`float` or iterable
        input number

    Returns
    -------
    :obj:`int` or :obj:`ndarray` containing ints
        exponent of input number(s)

    Example
    -------
    >>> from pyaerocom.mathutils import exponent
    >>> exponent(2340)
    3
    """
    return np.floor(np.log10(abs(np.asarray(num)))).astype(int)


def range_magnitude(low, high):
    """Returns magnitude of value range

    Parameters
    ----------
    low : float
        lower end of range
    high : float
        upper end of range

    Returns
    -------
    int
        magnitudes spanned by input numbers

    Example
    -------

    >>> range_magnitude(0.1, 100)
    3
    >>> range_magnitude(100, 0.1)
    -3
    >>> range_magnitude(1e-3, 1e6)
    9

    """
    return exponent(high) - exponent(low)


def estimate_value_range(vmin, vmax, extend_percent=0):
    """
    Round and extend input range to estimate lower and upper bounds of range

    Parameters
    ----------
    vmin : float
        lower value of range
    vmax : float
        upper value of range
    extend_percent : int
        percentage specifying to which extent the input range is supposed to be
        extended.

    Returns
    -------
    float
        estimated lower end of range
    float
        estimated upper end of range


    """
    if not vmax > vmin:
        raise ValueError("vmax needs to exceed vmin")
    # extent value range by +/- 5%
    offs = (vmax - vmin) * extend_percent * 0.01
    vmin, vmax = vmin - offs, vmax + offs

    if vmin != 0:
        exp = float(exponent(vmin))
    else:
        exp = float(exponent(vmax))
    # round values
    vmin = np.floor(vmin * 10 ** (-exp)) * 10.0 ** (exp)
    vmax = np.ceil(vmax * 10 ** (-exp)) * 10.0 ** (exp)
    return vmin, vmax



import numpy as np

from pyaerocom import region
from pyaerocom.config import ALL_REGION_NAME


class Filter:
    """Class that can be used to filter gridded and ungridded data objects

    Note
    ----
    - BETA version (currently being tested)
    - Can only filter spatially
    - Might be renamed to RegionFilter at some point in the future

    Todo
    ----
    Include also temporal filtering and other filter options (e.g. variable,
    etc.)
    """

    #: dictionary specifying altitude filters
    ALTITUDE_FILTERS = {
        "wMOUNTAINS": None,  # reserve namespace for
        "noMOUNTAINS": [-1e6, 1e3],
    }  # 1000 m upper limit

    LAND_OCN_FILTERS = ["LAND", "OCN"]  # these are HTAP filters

    NO_REGION_FILTER_NAME = ALL_REGION_NAME
    NO_ALTITUDE_FILTER_NAME = "wMOUNTAINS"

    _DELIM = "-"

    def __init__(self, name=None, region=None, altitude_filter=None, land_ocn=None, **kwargs):
        # default name (i.e. corresponds to no filtering)
        self._name = None

        # this will be used to store instance of Region associated with filter
        self._region = None
        if name is not None:
            self.name = name
        else:
            self.name = f"{self.NO_REGION_FILTER_NAME}-{self.NO_ALTITUDE_FILTER_NAME}"

    @property
    def name(self):
        """Name of filter

        String containing up to 3 substrings (delimited using dash -)
        containing: <region_id>-<altitude_filter>-<land_or_sea_only_info>
        """
        return self._name

    @name.setter
    def name(self, val):
        self._name = self._check_name_valid(val)

    def _check_name_valid(self, val):
        if not isinstance(val, list):
            if not isinstance(val, str):
                raise ValueError(f"Need list or string as input for name attr got {val}")
            spl = val.split(self._DELIM)
        else:
            spl = val
        # make sure there are no duplicate strings in the name
        spl = list(np.unique(spl))
        if len(spl) > 3:
            raise ValueError("Filter name must not exceed 3 specifications")
        reg = None
        alt_filter = None
        landsea = None
        for entry in spl:
            if entry in self.valid_regions:
                if entry in self.LAND_OCN_FILTERS:
                    if landsea is not None:
                        raise ValueError("Filter name must only contain one landsea specification")
                    landsea = entry
                else:
                    if reg is not None:
                        raise ValueError("Only one region may be specified")
                    reg = entry

            elif entry in self.valid_alt_filter_codes:
                if alt_filter is not None:
                    raise ValueError("Only one altitude filter can be specified")
                alt_filter = entry
            else:
                raise ValueError(f"Invalid input for filter name {entry}")
        if reg is None:
            reg = ALL_REGION_NAME
        if alt_filter is None:
            alt_filter = "wMOUNTAINS"

        lst = [reg, alt_filter]
        if landsea is not None:
            lst.append(landsea)
        return f"{self._DELIM}".join(lst)

    @property
    def spl(self):
        return self._name.split(self._DELIM)

    @property
    def region_name(self):
        """Name of region"""
        return self.spl[0]

    @property
    def region(self):
        """Region associated with this filter (instance of :class:`Region`)"""
        r = self._region
        if not isinstance(r, region.Region) or not r.name == self.region_name:
            self._region = region.Region(self.region_name)
        return self._region

    @property
    def lon_range(self):
        """Longitude range of region"""
        return self.region.lon_range

    @property
    def lat_range(self):
        """Latitude range of region"""
        return self.region.lat_range

    @property
    def alt_range(self):
        """Altitude range of filter"""
        return self.ALTITUDE_FILTERS[self.spl[1]]

    def from_list(self, lst):
        """Set filter name based on input list"""
        if not isinstance(lst, list):
            raise TypeError("Invalid input, need list...")
        if len(lst) > 3:
            raise ValueError(
                f"Maximum length 3 of individual filter entries exceeded for input {lst}"
            )
        self.name = "-".join(lst)

    @property
    def valid_alt_filter_codes(self):
        """Valid codes for altitude filters"""
        return list(self.ALTITUDE_FILTERS)

    @property
    def valid_land_sea_filter_codes(self):
        """Codes specifying land/sea filters"""
        return self.LAND_OCN_FILTERS

    @property
    def valid_regions(self):
        """Names of valid regions (AeroCom regions and HTAP regions)"""
        return region.all()

    @property
    def land_ocn(self):
        return None if len(self.spl) < 3 else self.spl[2]

    def to_dict(self):
        """Convert filter to dictionary"""
        return {
            "region": self.region_name,
            "lon_range": self.lon_range,
            "lat_range": self.lat_range,
            "alt_range": self.alt_range,
            "land_sea": self.land_ocn,
        }

    def apply(self, data_obj):
        """Apply filter to data object

        Parameters
        ----------
        data_obj : :obj:`UngriddedData`, :obj:`GriddedData`
            input data object that is supposed to be filtered

        Returns
        -------
        :obj:`UngriddedData`, :obj:`GriddedData`
            filtered data object

        Raises
        ------
        IOError
            if input is invalid
        """
        spl = self.spl

        if spl[0] != self.NO_REGION_FILTER_NAME:
            data_obj = data_obj.filter_region(spl[0])
        if spl[1] != self.NO_ALTITUDE_FILTER_NAME:
            alt_range = self.ALTITUDE_FILTERS[spl[1]]
            data_obj = data_obj.filter_altitude(alt_range)
        if len(spl) > 2:
            data_obj = data_obj.filter_region(spl[2])
        return data_obj

    def __call__(self, data_obj):
        return self.apply(data_obj)



import logging

import netCDF4

# avoid a broken autoadjusting chunk-cache in netcdf by setting a static cache
# see https://github.com/Unidata/netcdf-c/issues/2913
nc_cache = netCDF4.get_chunk_cache()
netCDF4.set_chunk_cache(2 * nc_cache[0], nc_cache[1], nc_cache[2])
logging.getLogger(__name__).debug("netcdf-chunk-cache fix applied")



import glob
import logging
import os
import pathlib
import shutil
from collections import namedtuple

import aerovaldb

from pyaerocom import const
from pyaerocom._lowlevel_helpers import (
    DirLoc,
    StrType,
    TypeValidator,
    sort_dict_by_name,
)
from pyaerocom.aeroval.collections import ObsCollection
from pyaerocom.aeroval.glob_defaults import (
    VariableInfo,
    extended_statistics,
    statistics_defaults,
    statistics_mean_trend,
    statistics_median_trend,
    statistics_model_only,
    statistics_obs_only,
    statistics_trend,
)
from pyaerocom.aeroval.json_utils import round_floats
from pyaerocom.aeroval.modelentry import ModelEntry
from pyaerocom.aeroval import EvalSetup
from pyaerocom.aeroval.varinfo_web import VarinfoWeb
from pyaerocom.colocation.colocated_data import ColocatedData
from pyaerocom.exceptions import EntryNotAvailable, VariableDefinitionError
from pyaerocom.stats.mda8.const import MDA8_OUTPUT_VARS
from pyaerocom.stats.stats import _init_stats_dummy
from pyaerocom.utils import recursive_defaultdict
from pyaerocom.variable_helpers import get_aliases, get_variable

MapInfo = namedtuple(
    "MapInfo",
    ["obs_network", "obs_var", "vert_code", "mod_name", "mod_var", "time_period"],
)

logger = logging.getLogger(__name__)


class ProjectOutput:
    """JSON output for project"""

    proj_id = StrType()

    json_basedir = DirLoc(assert_exists=True)

    def __init__(self, proj_id: str, resource: str | pathlib.Path | aerovaldb.AerovalDB):
        self.proj_id = proj_id

        if isinstance(resource, pathlib.Path):
            resource = str(resource)

        if isinstance(resource, str):
            self.avdb = aerovaldb.open(resource)
        elif isinstance(resource, aerovaldb.AerovalDB):
            self.avdb = resource
        else:
            raise ValueError(f"Expected string or AerovalDB, got {type(resource)}.")

        # TODO: Only works for json_files, check if this is needed, and remove / rewrite
        # functionality that requires direct knowledge of _basedir.
        self.json_basedir = self.avdb._basedir

    @property
    def proj_dir(self) -> str:
        """Project directory"""
        fp = os.path.join(self.json_basedir, self.proj_id)
        if not os.path.exists(fp):
            os.mkdir(fp)
            logger.info(f"Creating AeroVal project directory at {fp}")
        return fp

    @property
    def experiments_file(self) -> str:
        """json file containing region specifications"""
        fp = os.path.join(self.proj_dir, "experiments.json")
        return fp

    @property
    def available_experiments(self) -> list:
        """
        List of available experiments
        """
        return list(self.avdb.get_experiments(self.proj_id, default={}))


class ExperimentOutput(ProjectOutput):
    """JSON output for experiment"""

    cfg = TypeValidator(EvalSetup)

    def __init__(self, cfg: EvalSetup):
        self.cfg = cfg
        super().__init__(
            cfg.proj_id,
            (
                cfg.path_manager.json_basedir
                if cfg.path_manager.avdb_resource is None
                else cfg.path_manager.json_basedir
            ),
        )

        # dictionary that will be filled by json cleanup methods to check for
        # invalid or outdated json files across different output directories
        self._invalid = dict(models=[], obs=[])

    @property
    def exp_id(self) -> str:
        """Experiment ID"""
        return self.cfg.exp_id

    @property
    def exp_dir(self) -> str:
        """Experiment directory"""
        fp = os.path.join(self.proj_dir, self.exp_id)
        if not os.path.exists(fp):
            os.mkdir(fp)
            logger.info(f"Creating AeroVal experiment directory at {fp}")
        return fp

    @property
    def regions_file(self) -> str:
        """json file containing region specifications"""
        fp = os.path.join(self.exp_dir, "regions.json")
        return fp

    @property
    def statistics_file(self) -> str:
        """json file containing region specifications"""
        fp = os.path.join(self.exp_dir, "statistics.json")
        return fp

    @property
    def var_ranges_file(self) -> str:
        """json file containing region specifications"""
        fp = os.path.join(self.exp_dir, "ranges.json")
        return fp

    @property
    def menu_file(self) -> str:
        """json file containing region specifications"""
        fp = os.path.join(self.exp_dir, "menu.json")
        return fp

    @property
    def results_available(self) -> bool:
        """
        bool: True if results are available for this experiment, else False
        """
        if self.exp_id not in os.listdir(self.proj_dir):
            return False
        elif not len(self._get_json_output_files("map")) > 0:
            return False
        return True

    @property
    def out_dirs_json(self) -> dict:
        """
        json output directories (`dict`)
        """
        return self.cfg.path_manager.get_json_output_dirs()

    def update_menu(self) -> None:
        """Update menu

        The menu.json file is created based on the available json map files in the
        map directory of an experiment.

        Parameters
        ----------
        menu_file : str
            path to json menu file
        delete_mode : bool
            if True, then no attempts are being made to find json files for the
            experiment specified in `config`.

        """
        avail = self._create_menu_dict()
        avail = self._sort_menu_entries(avail)
        with self.avdb.lock():
            self.avdb.put_menu(avail, self.proj_id, self.exp_id)

    def update_interface(self) -> None:
        """
        Update web interface

        Steps:

        1. Check if results are available, and if so:
        2. Add entry for this experiment in experiments.json
        3. Create/update ranges.json file in experiment directory
        4. Update menu.json against available output and evaluation setup
        5. Synchronise content of heatmap json files with menu
        6. Create/update file statistics.json in experiment directory
        7. Copy json version of EvalSetup into experiment directory

        Returns
        -------
        None

        """
        if not self.results_available:
            logger.warning(f"no output available for experiment {self.exp_id} in {self.proj_id}")
            return
        exp_data = {"public": self.cfg.exp_info.public}
        self._add_entry_experiments_json(self.exp_id, exp_data)
        self._create_var_ranges_json()
        self.update_menu()
        self._sync_heatmaps_with_menu_and_regions()

        self._create_statistics_json()
        # AeroVal frontend needs periods to be set in config json file...
        # make sure they are
        self.cfg._check_time_config()
        with self.avdb.lock():
            self.avdb.put_config(self.cfg.json_repr(), self.proj_id, self.exp_id)

    def _sync_heatmaps_with_menu_and_regions(self) -> None:
        """
        Synchronise content of heatmap json files with content of menu.json
        """
        with self.avdb.lock():
            menu = self.avdb.get_menu(self.proj_id, self.exp_id, default={})
            all_regions = self.avdb.get_regions(self.proj_id, self.exp_id, default={})
            for uri in self.avdb.list_glob_stats(
                self.proj_id, self.exp_id, access_type=aerovaldb.AccessType.URI
            ):
                data = self.avdb.get_by_uri(uri)
                hm = {}
                for vardisp, info in menu.items():
                    obs_dict = info["obs"]
                    if vardisp not in hm:
                        hm[vardisp] = {}
                    for obs, vdict in obs_dict.items():
                        if obs not in hm[vardisp]:
                            hm[vardisp][obs] = {}
                        for vert_code, mdict in vdict.items():
                            if vert_code not in hm[vardisp][obs]:
                                hm[vardisp][obs][vert_code] = {}
                            for mod, minfo in mdict.items():
                                if mod not in hm[vardisp][obs][vert_code]:
                                    hm[vardisp][obs][vert_code][mod] = {}
                                modvar = minfo["model_var"]
                                hm_data = data[vardisp][obs][vert_code][mod][modvar]
                                hm_data = self._check_hm_all_regions_avail(all_regions, hm_data)
                                hm[vardisp][obs][vert_code][mod][modvar] = hm_data

                self.avdb.put_by_uri(hm, uri)

    def _check_hm_all_regions_avail(self, all_regions, hm_data) -> dict:
        if all([x in hm_data for x in all_regions]):
            return hm_data
        # some regions are not available in this subset
        periods = self.cfg.time_cfg._get_all_period_strings()
        dummy_stats = _init_stats_dummy()
        for region in all_regions:
            if region not in hm_data:
                hm_data[region] = {}
                for per in periods:
                    hm_data[region][per] = dummy_stats
        return hm_data

    @staticmethod
    def _info_from_map_file(filename: str) -> MapInfo:
        """
        Separate map filename into meta info on obs and model content

        Parameters
        ----------
        filename : str
            name of file in "map" subdirectory of json output directory for
            this experiment

        Raises
        ------
        ValueError
            if input filename is invalid

        Returns
        -------
        str
            name of observation network
        str
            name of observation variable
        str
            name of vertical code (e.g. Surface)
        str
            name of model
        str
            name of model variable
        str
            Time period
        """
        spl = os.path.basename(filename).split(".json")[0].split("_")
        if len(spl) != 4:
            raise ValueError(
                f"invalid map filename: {filename}. Must "
                f"contain exactly 3 underscores _ to separate "
                f"obsinfo, vertical, model info, and periods"
            )
        obsinfo = spl[0]
        vert_code = spl[1]
        modinfo = spl[2]
        time_period = spl[3]

        mspl = modinfo.split("-")
        mod_var = mspl[-1]
        mod_id = "-".join(mspl[:-1])

        ospl = obsinfo.split("-")
        obs_var = ospl[-1]
        obs_network = "-".join(ospl[:-1])

        return MapInfo(obs_network, obs_var, vert_code, mod_id, mod_var, time_period)

    def _results_summary(self) -> dict[str, list[str]]:
        res = [[], [], [], [], [], []]
        files = self._get_json_output_files("map")
        for file in files:
            map_info = self._info_from_map_file(file)
            for i, entry in enumerate(map_info):
                res[i].append(entry)
        output = {}
        for i, name in enumerate(["obs", "ovar", "vc", "mod", "mvar", "per"]):
            output[name] = list(set(res[i]))
        return output

    def clean_json_files(self) -> list[str]:
        """Checks all existing json files and removes outdated data

        This may be relevant when updating a model name or similar.

        Returns:
        list[str] :
            The list of file paths that where modified / removed.
        """
        modified = []
        logger.info(
            "Running clean_json_files: Checking json output directories for "
            "outdated or invalid data and cleaning up."
        )
        outdirs = self.out_dirs_json
        mapfiles = self._get_json_output_files("map")
        rmmap = []
        vert_codes = self.cfg.obs_cfg.all_vert_types
        for file_path in mapfiles:
            try:
                (
                    obs_network,
                    obs_var,
                    vert_code,
                    mod_name,
                    mod_var,
                    time_period,
                ) = self._info_from_map_file(file_path)
            except Exception as e:
                logger.warning(
                    f"FATAL: invalid file convention for map json file:"
                    f" {file_path}. This file will be deleted. Error message: "
                    f"{repr(e)}"
                )
                rmmap.append(file_path)
                continue
            if not self._is_part_of_experiment(obs_network, obs_var, mod_name, mod_var):
                rmmap.append(file_path)
            elif vert_code not in vert_codes:
                rmmap.append(file_path)

        scatfiles = os.listdir(outdirs["scat"])
        for file_path in rmmap:  # delete map files
            logger.info(f"Deleting outdated map json file: {file_path}.")
            os.remove(file_path)
            modified.append(file_path)
            fname = os.path.basename(file_path)
            if fname in scatfiles:
                scfp = os.path.join(outdirs["scat"], fname)
                logger.info(f"Deleting outdated scatter json file: {scfp}.")
                os.remove(scfp)
                modified.append(file_path)

        tsfiles = self._get_json_output_files("ts")

        for file_path in tsfiles:
            if self._check_clean_ts_file(file_path):
                modified.append(file_path)
        modified.extend(self._clean_modelmap_files())
        self.update_interface()  # will take care of heatmap data
        return modified

    def _check_clean_ts_file(self, fp) -> bool:
        fname = os.path.basename(fp)
        spl = fname.split(".json")[0].split("_")
        vert_code, obsinfo = spl[-1], spl[-2]
        if vert_code not in self.cfg.obs_cfg.all_vert_types:
            logger.warning(
                f"Invalid or outdated vert code {vert_code} in ts file {fp}. File will be deleted."
            )
            os.remove(fp)
            return True
        obs_name = str.join("-", obsinfo.split("-")[:-1])
        if obs_name in self._invalid["obs"]:
            logger.info(
                f"Invalid or outdated obs name {obs_name} in ts file {fp}. "
                f"File will be deleted."
            )
            os.remove(fp)
            return True

        with self.avdb.lock():
            try:
                # TODO: Hack to get uri. Ideally this should be rewritten to use URIs directly further
                # up.
                uri = self.avdb._get_uri_for_file(fp)
                data = self.avdb.get_by_uri(uri)
            except Exception:
                logger.exception(f"FATAL: detected corrupt json file: {fp}. Removing file...")
                os.remove(fp)
                return True

            models_avail = list(data)
            models_in_exp = self.cfg.model_cfg.web_interface_names
            if all([mod in models_in_exp for mod in models_avail]):
                # nothing to clean up
                return False
            modified = False
            data_new = {}
            for mod_name in models_avail:
                if mod_name in models_in_exp:
                    data_new[mod_name] = data[mod_name]
                else:
                    modified = True
                    logger.info(f"Removing data for model {mod_name} from ts file: {fp}")

            self.avdb.put_by_uri(data_new, uri)
        return modified

    def _clean_modelmap_files(self) -> list[str]:
        # Note: to be called after cleanup of files in map subdir
        json_files = self._get_json_output_files("contour")
        rm = []
        for file in json_files:
            if not self.cfg.webdisp_opts.add_model_maps:
                rm.append(file)
            else:
                fname = os.path.basename(file)
                spl = fname.split(".")[0].split("_")
                if not len(spl) == 2:
                    msg = f"FATAL: invalid file convention for map json file: {file}."
                    if len(spl) > 2:
                        msg += "Likely due to underscore being present in model or variable name."
                    rm.append(file)
                    logger.warning(msg)
                elif spl[-1] in self._invalid["models"]:
                    rm.append(file)

        removed = []
        for file in rm:
            os.remove(file)
            removed.append(file)
            file1 = file.replace(".json", ".geojson")
            if os.path.exists(file1):
                os.remove(file1)
                removed.append(file)
        return removed

    def delete_experiment_data(self, also_coldata=True) -> None:
        """Delete all data associated with a certain experiment

        Note
        ----
        This simply deletes the experiment directory with all the json files
        and, if `also_coldata` is True, also the associated co-located data
        objects.

        Parameters
        ----------
        also_coldata : bool
            if True and if output directory for colocated data is default and
            specific for input experiment ID, then also all associated colocated
            NetCDF files are deleted. Defaults to True.
        """
        self.avdb.rm_experiment_data(self.proj_id, self.exp_id)

        if also_coldata:
            coldir = self.cfg.path_manager.get_coldata_dir()
            if os.path.exists(coldir):
                logger.info(f"Deleting everything under {coldir}")
                shutil.rmtree(coldir)
        self._del_entry_experiments_json(self.exp_id)

    def get_model_order_menu(self) -> list:
        """Order of models in menu

        Note
        ----
        Returns empty list if no specific order is to be used in which case
        the models will be alphabetically ordered
        """
        order = []
        if len(self.cfg.webdisp_opts.model_order_menu) > 0:
            if self.cfg.webdisp_opts.modelorder_from_config:
                raise AttributeError(
                    "Conflict: modelorder_from_config must be deactivated if "
                    "model_order_menu is specified explicitly"
                )
            order.extend(self.cfg.webdisp_opts.model_order_menu)
        elif self.cfg.webdisp_opts.obsorder_from_config:
            order.extend(self.cfg.model_cfg.web_interface_names)
        return order

    def get_obs_order_menu(self) -> list:
        """Order of observation entries in menu"""
        order = []
        if len(self.cfg.webdisp_opts.obs_order_menu) > 0:
            if self.cfg.webdisp_opts.obsorder_from_config:
                raise AttributeError(
                    "Conflict: obsorder_from_config must be deactivated if "
                    "obs_order_menu is specified explicitly"
                )
            order.extend(self.cfg.webdisp_opts.obs_order_menu)
        elif self.cfg.webdisp_opts.obsorder_from_config:
            order.extend(self.cfg.obs_cfg.web_interface_names)
        return order

    def _get_json_output_files(self, dirname) -> list[str]:
        dirloc = self.out_dirs_json[dirname]
        return glob.glob(f"{dirloc}/*.json")

    def _get_cmap_info(self, var) -> dict[str, str | list[float]]:
        var_ranges_defaults = self.cfg.var_scale_colmap
        if var in var_ranges_defaults:
            return var_ranges_defaults[var]
        try:
            varinfo = VarinfoWeb(var)
            # TODO: get unit from pyaerocom/data/variables.ini
            info = dict(scale=varinfo.cmap_bins, colmap=varinfo.cmap, unit=varinfo.unit)
        except (VariableDefinitionError, AttributeError):
            info = var_ranges_defaults["default"]
            logger.warning(
                f"Failed to infer cmap and variable "
                f"ranges for {var}, using default "
                f"settings which are {info}"
            )

        return info

    def _create_var_ranges_json(self) -> None:
        with self.avdb.lock():
            ranges = self.avdb.get_ranges(self.proj_id, self.exp_id, default={})

            avail = self._results_summary()
            all_vars = list(set(avail["ovar"] + avail["mvar"]))
            for var in all_vars:
                if var not in ranges or ranges[var]["scale"] == []:
                    ranges[var] = self._get_cmap_info(var)
                ranges[var]["unit"] = get_variable(var).units
            self.avdb.put_ranges(ranges, self.proj_id, self.exp_id)

    def _create_statistics_json(self) -> None:
        if self.cfg.statistics_opts.obs_only_stats:
            stats_info = statistics_obs_only
        elif self.cfg.statistics_opts.model_only_stats:
            stats_info = statistics_model_only
        else:
            stats_info = statistics_defaults
            stats_info.update(extended_statistics)

        # configurable statistics - drop any statistics provided in drop_stats
        if self.cfg.statistics_opts.drop_stats:
            for stat in self.cfg.statistics_opts.drop_stats:
                stats_info.pop(stat, None)

        # configure the number of decimals shown in statistics if provided
        if self.cfg.statistics_opts.stats_decimals:
            for stat in stats_info:
                stats_info[stat].update(decimals=self.cfg.statistics_opts.stats_decimals)

        if self.cfg.statistics_opts.add_trends:
            if self.cfg.processing_opts.obs_only:
                obs_statistics_trend = {
                    key: val for key, val in statistics_trend.items() if "mod" not in key
                }
                stats_info.update(obs_statistics_trend)
            else:
                stats_info.update(statistics_trend)

                if self.cfg.statistics_opts.avg_over_trends:
                    stats_info.update(statistics_mean_trend)
                    stats_info.update(statistics_median_trend)

        with self.avdb.lock():
            self.avdb.put_statistics(stats_info, self.proj_id, self.exp_id)

    def _get_var_name_and_type(self, var_name: str) -> VariableInfo:
        """Get menu name and type of observation variable

        Parameters
        ----------
        var_name : str
            Name of variable

        Returns
        -------
        VariableInfo :
            named tuple containing
            - menu name of this variable.
            - Vertical type of this variable (ie. 2D, 3D).
            - Category of this variable.
        """
        if var_name in self.cfg.var_web_info:
            name, tp, cat = self.cfg.var_web_info[var_name]
        else:
            name, tp, cat = var_name, "UNDEFINED", "UNDEFINED"
            logger.warning(f"Missing menu name definition for var {var_name}.")

        return VariableInfo(name, tp, cat)

    def _init_menu_entry(self, var: str) -> dict:
        name, tp, cat = self._get_var_name_and_type(var)
        out = {"type": tp, "cat": cat, "name": name, "obs": {}}
        try:
            lname = const.VARS[var].description
        except VariableDefinitionError:
            lname = "UNDEFINED"

        out["longname"] = lname
        try:
            # Comes in as a string. split() here breaks up based on space and returns either just the element in a list or the components of the string in a list
            only_use_in = const.VARS[var].only_use_in.split(" ")
            # only return only_use_in if key exists, otherwise do not
            out["only_use_in"] = only_use_in
        except AttributeError:
            pass
        return out

    def _check_ovar_mvar_entry(
        self, mcfg: ModelEntry, mod_var, ocfg: ObsCollection, obs_var
    ) -> bool:
        muv = mcfg.model_use_vars
        mrv = mcfg.model_rename_vars

        mvar_aliases = get_aliases(mod_var)
        for ovar, mvars in mcfg.model_add_vars.items():
            if obs_var in mvars:
                # for evaluation of entries in model_add_vars, the output json
                # files use the model variable both for obs and for model as a
                # workaround for the AeroVal heatmap display (which is based on
                # observation variables on the y-axis). E.g. if
                # model_add_vars=dict(od550aer=['od550so4']) then there will
                # 2 co-located data objects one where model od550aer is
                # co-located with obs od550aer and one where model od550so4
                # is co-located with obs od550aer, thus for the latter,
                # the obs variable is set to od550so4, so it shows up as a
                # separate entry in AeroVal.
                if obs_var in mrv:
                    # model_rename_vars is specified for that obs variable,
                    # e.g. using the above example, the output obs variable
                    # would be od550so4, however, the user want to rename
                    # the corresponding model variable via e.g.
                    # model_rename_vars=dict(od550so4='MyVar'). Thus,
                    # check if model variable is MyVar
                    if mod_var == mrv[obs_var]:
                        return True
                elif obs_var == mod_var:
                    # if match, then they should be the same here
                    return True

        obs_vars = ocfg.get_all_vars()
        if obs_var in obs_vars:
            if obs_var in muv:
                mvar_to_use = muv[obs_var]
                if mvar_to_use == mod_var:
                    # obs var is different from mod_var but this mapping is
                    # specified in mcfg.model_use_vars
                    return True
                elif mvar_to_use in mvar_aliases:
                    # user specified an alias name in config for the
                    # observation variable e.g. model_use_vars=dict(
                    # ac550aer=absc550dryaer).
                    return True
                elif mvar_to_use in mrv and mrv[mvar_to_use] == mod_var:
                    # user wants to rename the model variable
                    return True
            if obs_var in mrv and mrv[obs_var] == mod_var:
                # obs variable is in model_rename_vars
                return True
            elif mod_var in get_aliases(obs_var):
                # model var is an alias to obs var e.g. sconcpm10 to concpm10
                return True
            elif mod_var == obs_var:
                # default setting, includes cases where mcfg.model_use_vars
                # is set and the value of the model variable in
                # mcfg.model_use_vars is an alias for obs_var
                return True
        return False

    def _is_part_of_experiment(self, obs_name, obs_var, mod_name, mod_var) -> bool:
        """
        Check if input combination of model and obs var is valid

        Note
        ----
        The input parameters are supposed to be retrieved from json files
        stored in the map subdirectory of an existing AeroVal experiment. In
        complex setup cases the variable mapping (model / obs variables)
        used in these json filenames may not be the trivial one expected from
        the configuaration. These are cases where one specifies
        model_add_vars, or model_use_vars or model_rename_vars in a model
        entry.

        Parameters
        ----------
        obs_name : str
            Name of obs dataset.
        obs_var : str
            Name of obs variable.
        mod_name : str
            Name of model
        mod_var : str
            Name of model variable

        Returns
        -------
        bool
            True if this combination is valid, else False.

        """

        # MDA8 is computed on-the-fly ONLY if a MDA8_INPUT_VAR at hourly freq is detected.
        # Consequently, it is not specified in a config but should be included as part of the experiment.
        if obs_var in MDA8_OUTPUT_VARS and mod_var in MDA8_OUTPUT_VARS:
            return True

        # get model entry for model name
        try:
            mcfg = self.cfg.model_cfg.get_entry(mod_name)
        except EntryNotAvailable:
            self._invalid["models"].append(mod_name)
            return False
        # mapping of obs / model variables to be used

        # search obs entry (may have web_interface_name set, so have to
        # check keys of ObsCollection but also the individual entries for
        # occurence of web_interface_name).
        allobs = self.cfg.obs_cfg
        obs_matches = []
        for ocfg in allobs:
            if obs_name == allobs.get_web_interface_name(ocfg.obs_name):
                obs_matches.append(ocfg)
        if len(obs_matches) == 0:
            self._invalid["obs"].append(obs_name)
            # obs dataset is not part of experiment
            return False
        # first, check model_add_vars
        for ocfg in obs_matches:
            if self._check_ovar_mvar_entry(mcfg, mod_var, ocfg, obs_var):
                return True
        return False

    def _create_menu_dict(self) -> dict:
        new = {}
        files = self._get_json_output_files("map")
        for file in files:
            (obs_name, obs_var, vert_code, mod_name, mod_var, per) = self._info_from_map_file(file)

            if self._is_part_of_experiment(obs_name, obs_var, mod_name, mod_var):
                mcfg = self.cfg.model_cfg.get_entry(mod_name)
                var = mcfg.get_varname_web(mod_var, obs_var)
                if var not in new:
                    new[var] = self._init_menu_entry(var)

                if obs_name not in new[var]["obs"]:
                    new[var]["obs"][obs_name] = {}

                if vert_code not in new[var]["obs"][obs_name]:
                    new[var]["obs"][obs_name][vert_code] = {}
                if mod_name not in new[var]["obs"][obs_name][vert_code]:
                    new[var]["obs"][obs_name][vert_code][mod_name] = {}

                new[var]["obs"][obs_name][vert_code][mod_name] = {
                    "model_id": mcfg.model_id,
                    "model_var": mod_var,
                    "obs_var": obs_var,
                }
            else:
                logger.warning(
                    f"Invalid entry: model {mod_name} ({mod_var}), obs {obs_name} ({obs_var})"
                )
        return new

    def _sort_menu_entries(self, avail: dict) -> dict:
        """
        Used in method :func:`update_menu_evaluation_iface`

        Sorts results of different menu entries (i.e. variables, observations
        and models).

        Parameters
        ----------
        avail : dict
            nested dictionary contining info about available results

        Returns
        -------
        dict
            input dictionary sorted in variable, obs and model layers. The order
            of variables, observations and models may be specified in
            AerocomEvaluation class and if not, alphabetic order is used.

        """
        # sort first layer (i.e. variables)
        avail = sort_dict_by_name(avail, pref_list=self.cfg.webdisp_opts.var_order_menu)

        new_sorted = {}
        for var, info in avail.items():
            new_sorted[var] = info
            obs_order = self.get_obs_order_menu()
            sorted_obs = sort_dict_by_name(info["obs"], pref_list=obs_order)
            new_sorted[var]["obs"] = sorted_obs
            for obs_name, vert_codes in sorted_obs.items():
                vert_codes_sorted = sort_dict_by_name(vert_codes)
                new_sorted[var]["obs"][obs_name] = vert_codes_sorted
                for vert_code, models in vert_codes_sorted.items():
                    model_order = self.get_model_order_menu()
                    models_sorted = sort_dict_by_name(models, pref_list=model_order)
                    new_sorted[var]["obs"][obs_name][vert_code] = models_sorted
        return new_sorted

    def _add_entry_experiments_json(self, exp_id: str, data) -> None:
        with self.avdb.lock():
            current = self.avdb.get_experiments(self.proj_id, default={})

            current[exp_id] = data

            self.avdb.put_experiments(current, self.proj_id)

    def _del_entry_experiments_json(self, exp_id) -> None:
        """
        Remove an entry from experiments.json

        Parameters
        ----------
        exp_id : str
            name of experiment

        Returns
        -------
        None

        """
        with self.avdb.lock():
            current = self.avdb.get_experiments(self.proj_id, default={})

            try:
                del current[exp_id]
            except KeyError:
                logger.warning(f"no such experiment registered: {exp_id}")
            self.avdb.put_experiments(current, self.proj_id)

    def reorder_experiments(self, exp_order=None) -> None:
        """Reorder experiment order in evaluation interface

        Puts experiment list into order as specified by `exp_order`, all
        remaining experiments are sorted alphabetically.

        Parameters
        ----------
        exp_order : list, optional
            desired experiment order, if None, then alphabetical order is used.
        """
        if exp_order is None:
            exp_order = []
        elif not isinstance(exp_order, list):
            raise ValueError("need list as input")

        with self.avdb.lock():
            current = self.avdb.get_experiments(self.proj_id, default={})

            current = sort_dict_by_name(current, pref_list=exp_order)
            self.avdb.put_experiments(current, self.proj_id)

    def add_heatmap_timeseries_entry(
        self,
        entry: dict,
        region: str,
        network: str,
        obsvar: str,
        layer: str,
        modelname: str,
        modvar: str,
    ):
        """Adds a heatmap entry to hm/ts

        :param entry: The entry to be added.
        :param network: Observation network
        :param obsvar: Observation variable
        :param layer: Vertical layer
        :param modelname: Model name
        :param modvar: Model variable
        """
        project = self.proj_id
        experiment = self.exp_id

        with self.avdb.lock():
            glob_stats = self.avdb.get_heatmap_timeseries(
                project, experiment, region, network, obsvar, layer, default={}
            )
            glob_stats = recursive_defaultdict(glob_stats)
            glob_stats[obsvar][network][layer][modelname][modvar] = round_floats(entry)
            self.avdb.put_heatmap_timeseries(
                glob_stats,
                project,
                experiment,
                region,
                network,
                obsvar,
                layer,
            )

    def add_forecast_entry(
        self,
        entry: dict,
        region: str,
        network: str,
        obsvar: str,
        layer: str,
        modelname: str,
        modvar: str,
    ):
        """Adds a forecast entry to forecast

        :param entry: The entry to be added.
        :param network: Observation network
        :param obsvar: Observation variable
        :param layer: Vertical layer
        :param modelname: Model name
        :param modvar: Model variable
        """
        project = self.proj_id
        experiment = self.exp_id

        with self.avdb.lock():
            glob_stats = self.avdb.get_forecast(
                project, experiment, region, network, obsvar, layer, default={}
            )
            glob_stats = recursive_defaultdict(glob_stats)
            glob_stats[obsvar][network][layer][modelname][modvar] = round_floats(entry)
            self.avdb.put_forecast(
                glob_stats,
                project,
                experiment,
                region,
                network,
                obsvar,
                layer,
            )

    def add_heatmap_entry(
        self,
        entry,
        frequency: str,
        network: str,
        obsvar: str,
        layer: str,
        modelname: str,
        modvar: str,
    ):
        """Adds a heatmap entry to glob_stats

        :param entry: The entry to be added.
        :param region: The region (eg. ALL)
        :param obsvar: Observation variable.
        :param layer: Vertical Layer (eg. SURFACE)
        :param modelname: Model name
        :param modelvar: Model variable.
        """
        project = self.proj_id
        experiment = self.exp_id

        with self.avdb.lock():
            glob_stats = self.avdb.get_glob_stats(project, experiment, frequency, default={})
            glob_stats = recursive_defaultdict(glob_stats)
            glob_stats[obsvar][network][layer][modelname][modvar] = entry
            self.avdb.put_glob_stats(glob_stats, project, experiment, frequency)

    def write_station_data(self, data):
        """Writes timeseries weekly.

        :param data: Data to be written.
        """
        project = self.proj_id
        experiment = self.exp_id

        location = data["station_name"]
        network = data["obs_name"]
        obsvar = data["var_name_web"]
        layer = data["vert_code"]
        modelname = data["model_name"]
        with self.avdb.lock():
            station_data = self.avdb.get_timeseries_weekly(
                project, experiment, location, network, obsvar, layer, default={}
            )
            station_data[modelname] = round_floats(data)
            self.avdb.put_timeseries_weekly(
                station_data, project, experiment, location, network, obsvar, layer
            )

    def write_timeseries(self, data):
        """Write timeseries

        Args:
            data: The timeseries object to be written.

        Note:
        -----
        All necessary metadata will be read from the data object.
        """
        if not isinstance(data, list):
            data = [data]

        project = self.proj_id
        experiment = self.exp_id
        with self.avdb.lock():
            for d in data:
                location = d["station_name"]
                network = d["obs_name"]
                obsvar = d["var_name_web"]
                layer = d["vert_code"]
                modelname = d["model_name"]

                timeseries = self.avdb.get_timeseries(
                    project, experiment, location, network, obsvar, layer, default={}
                )
                timeseries[modelname] = round_floats(d)
                self.avdb.put_timeseries(
                    timeseries, project, experiment, location, network, obsvar, layer
                )

    def add_profile_entry(
        self,
        data: ColocatedData,
        profile_viz: dict,
        periods: list[str],
        seasons: list[str],
        location,
        network,
        obsvar,
    ):
        """Adds an entry for the colocated data to profiles.json.

        Args:
            data (ColocatedData): For this vertical layer
            profile_viz (dict): Output of process_profile_data()
            periods (list[str]): periods to compute over (years)
            seasons (list[str]): seasons to compute over (e.g., All, DJF, etc.)
        """
        with self.avdb.lock():
            current = self.avdb.get_profiles(
                self.proj_id, self.exp_id, location, network, obsvar, default={}
            )
            current = recursive_defaultdict(current)

            for freq, coldata in data.items():
                model_name = coldata.model_name

                midpoint = (
                    float(coldata.data.attrs["vertical_layer"]["end"])
                    + float(coldata.data.attrs["vertical_layer"]["start"])
                ) / 2
                if "z" not in current[model_name]:
                    current[model_name]["z"] = [midpoint]  # initalize with midpoint

                if (
                    midpoint > current[model_name]["z"][-1]
                ):  # only store incremental increases in the layers
                    current[model_name]["z"].append(midpoint)

                # old boilerplate to get around recursive_default_dict issues
                if "obs" not in current[model_name]:
                    current[model_name]["obs"] = {}

                if freq not in current[model_name]["obs"]:
                    current[model_name]["obs"][freq] = {}

                if "mod" not in current[model_name]:
                    current[model_name]["mod"] = {}

                if freq not in current[model_name]["mod"]:
                    current[model_name]["mod"][freq] = {}

                for per in periods:
                    for season in seasons:
                        perstr = f"{per}-{season}"

                        if perstr not in current[model_name]["obs"][freq]:
                            current[model_name]["obs"][freq][perstr] = []
                        if perstr not in current[model_name]["mod"][freq]:
                            current[model_name]["mod"][freq][perstr] = []

                        current[model_name]["obs"][freq][perstr].append(
                            profile_viz["obs"][freq][perstr]
                        )
                        current[model_name]["mod"][freq][perstr].append(
                            profile_viz["mod"][freq][perstr]
                        )

                if "metadata" not in current[model_name]:
                    current[model_name]["metadata"] = {
                        "z_unit": coldata.data.attrs["altitude_units"],
                        "z_description": "Altitude ASL",
                        "z_long_description": "Altitude Above Sea Level",
                        "unit": coldata.unitstr,
                    }
                current[model_name] = round_floats(current[model_name])

            self.avdb.put_profiles(current, self.proj_id, self.exp_id, location, network, obsvar)



"""
Interface for reading EEA AqERep files (formerly known as Airbase data).
"""

import glob
import gzip
import logging
import os
import pathlib
import shutil
import tempfile

import cf_units
import numpy as np
import pandas as pd
from tqdm import tqdm

from pyaerocom.exceptions import EEAv2FileError, TemporalResolutionError
from pyaerocom.io.helpers import get_country_name_from_iso
from pyaerocom.io.readungriddedbase import ReadUngriddedBase
from pyaerocom.stationdata import StationData
from pyaerocom.ungriddeddata import UngriddedData

logger = logging.getLogger(__name__)


class ReadEEAAQEREPBase(ReadUngriddedBase):
    """Class for reading EEA AQErep data

    Extended class derived from  low-level base class
    :class:`ReadUngriddedBase` that contains some more functionality.

    Note
    ----
    Currently only single variable reading into an :class:`UngriddedData`
    object is supported.
    """

    #: Mask for identifying datafiles
    _FILEMASK = "*.csv"

    #: Version log of this class (for caching)
    __version__ = "0.09"

    #: Column delimiter
    FILE_COL_DELIM = ","

    #: Name of the dataset (OBS_ID)
    DATA_ID = ""  # change this since we added more vars?

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [DATA_ID]

    #: There is no global ts_type but it is specified in the data files...
    TS_TYPE = "variable"

    #: sampling frequencies found in data files
    TS_TYPES_FILE = {
        "hour": "hourly",
        "day": "daily",
    }

    #: Dictionary specifying values corresponding to invalid measurements
    #: there's no value for NaNs in this data set. It uses an empty string
    NAN_VAL = {}

    #: Dictionary specifying the file column names (values) for each Aerocom
    #: variable (keys)
    # There's only one variable in each file named concentration
    VAR_NAMES_FILE = {}
    VAR_NAMES_FILE["concso2"] = "concentration"
    VAR_NAMES_FILE["conco3"] = "concentration"
    VAR_NAMES_FILE["concno2"] = "concentration"
    VAR_NAMES_FILE["concco"] = "concentration"
    VAR_NAMES_FILE["concno"] = "concentration"
    VAR_NAMES_FILE["concpm10"] = "concentration"
    VAR_NAMES_FILE["concpm25"] = "concentration"
    VAR_NAMES_FILE["vmro3"] = "concentration"
    VAR_NAMES_FILE["vmro3max"] = "concentration"
    VAR_NAMES_FILE["vmrno2"] = "concentration"

    VAR_NAMES_FILE["concSso2"] = "concentration"
    VAR_NAMES_FILE["concNno"] = "concentration"
    VAR_NAMES_FILE["concNno2"] = "concentration"

    #: units of variables in files (needs to be defined for each variable supported)
    VAR_UNITS_FILE = {
        "µg/m3": "ug m-3",
        "mg/m3": "mg m-3",
        "µgS/m3": "ug S m-3",
        "µgN/m3": "ug N m-3",
        "ppb": "ppb",
    }

    #: file masks for the data files
    FILE_MASKS = dict(
        concso2="**/??_1_*_timeseries.csv*",
        concpm10="**/??_5_*_timeseries.csv*",
        conco3="**/??_7_*_timeseries.csv*",
        vmro3="**/??_7_*_timeseries.csv*",
        vmro3max="**/??_7_*_timeseries.csv*",
        concno2="**/??_8_*_timeseries.csv*",
        vmrno2="**/??_8_*_timeseries.csv*",
        concco="**/??_10_*_timeseries.csv*",
        concno="**/??_38_*_timeseries.csv*",
        concpm25="**/??_6001_*_timeseries.csv*",
        concSso2="**/??_1_*_timeseries.csv*",
        concNno2="**/??_8_*_timeseries.csv*",
        concNno="**/??_38_*_timeseries.csv*",
    )

    # conversion factor between concX and vmrX
    CONV_FACTOR = {}

    CONV_FACTOR["concSso2"] = np.float64(0.50052292274792)
    CONV_FACTOR["concNno2"] = np.float64(0.3044517868011477)
    CONV_FACTOR["concNno"] = np.float64(0.466788868521913)
    CONV_FACTOR["vmro3"] = np.float64(
        0.493
    )  # retrieved using STD atmosphere from geonum and pya.mathutils.concx_to_vmrx
    CONV_FACTOR["vmro3max"] = np.float64(
        0.493
    )  # retrieved using STD atmosphere from geonum and pya.mathutils.concx_to_vmrx
    CONV_FACTOR["vmrno2"] = np.float64(
        0.514
    )  # retrieved using STD atmosphere from geonum and pya.mathutils.concx_to_vmrx

    # unit of the converted property after the conversion
    CONV_UNIT = {}
    CONV_UNIT["concSso2"] = "µgS/m3"
    CONV_UNIT["concNno2"] = "µgN/m3"
    CONV_UNIT["concNno"] = "µgN/m3"
    CONV_UNIT["vmro3"] = "ppb"
    CONV_UNIT["vmro3max"] = "ppb"
    CONV_UNIT["vmrno2"] = "ppb"

    #: field name of the start time of the measurement (in lower case)
    START_TIME_NAME = "datetimebegin"

    #: filed name of the end time of the measurement (in lower case)
    END_TIME_NAME = "datetimeend"

    #: dictionary that connects the EEA variable codes with aerocom variable names
    VAR_CODES = {}
    VAR_CODES["1"] = "concso2"
    VAR_CODES["5"] = "concpm10"
    VAR_CODES["7"] = "conco3"
    VAR_CODES["8"] = "concno2"
    VAR_CODES["10"] = "concco"
    VAR_CODES["38"] = "concno"
    VAR_CODES["6001"] = "concpm25"

    #: column name that holds the EEA variable code
    VAR_CODE_NAME = "airpollutantcode"

    #: List of variables that are provided by this dataset (will be extended
    #: by auxiliary variables on class init, for details see __init__ method of
    #: base class ReadUngriddedBase)
    PROVIDES_VARIABLES = list(VAR_NAMES_FILE)

    #: there's no general instrument name in the data
    INSTRUMENT_NAME = "unknown"

    #: max time steps to read per file (hourly data)
    # to make numpy array allocation size static
    MAX_LINES_TO_READ = 24 * 366

    #: file name of the metadata file
    #: this will be prepended with a data path later on
    # this file is in principe updated once a day.
    # so we night consider updating it from within the code later on.
    # URL: https://discomap.eea.europa.eu/map/fme/metadata/PanEuropean_metadata.csv
    DEFAULT_METADATA_FILE = "metadata.csv"

    #: Name of latitude variable in metadata file
    LATITUDENAME = "latitude"

    #: name of longitude variable in metadata file
    LONGITUDENAME = "longitude"

    #: name of altitude variable in metadata file
    ALTITUDENAME = "altitude"

    #: this class reads the European Environment Agency's Eionet data
    #: for details please read
    #: https://www.eea.europa.eu/about-us/countries-and-eionet
    WEBSITE = "https://discomap.eea.europa.eu/map/fme/AirQualityExport.htm"

    #: Eionet offers 2 data revisions
    #: E2a (near real time) and E1a (quality controlled)
    #: this class reads the E2a data for now.
    # But by changing the base path
    # and this constant, it can also read the E1a data set
    DATA_PRODUCT = ""

    AUX_REQUIRES = {
        "vmro3max": ["conco3"],
        "vmro3": ["conco3"],
        "vmrno2": ["concno2"],
        "concNno2": ["concno2"],
        "concNno": ["concno"],
        "concSso2": ["concso2"],
    }

    AUX_FUNS = {
        "vmro3": NotImplementedError(),
        "vmro3max": NotImplementedError(),
        "vmrno2": NotImplementedError(),
        "concNno2": NotImplementedError(),
        "concNno": NotImplementedError(),
        "concSso2": NotImplementedError(),
    }

    def __init__(self, data_id=None, data_dir=None):
        super().__init__(data_id=data_id, data_dir=data_dir)
        self._metadata = None

    @property
    def DEFAULT_VARS(self):
        """List of default variables"""
        return [self.VAR_CODES["7"]]

    @property
    def DATASET_NAME(self):
        """Name of the dataset"""
        return self.data_id

    def read_file(self, filename, var_name, vars_as_series=False):
        """Read a single EEA file

        Note that there's only a single variable in the file

        Parameters
        ----------
        filename : str
            Absolute path to filename to read.
        var_name : str
            Name of variable in file.
        vars_as_series : bool
            If True, the data columns of all variables in the result dictionary
            are converted into pandas Series objects.

        Returns
        -------
        StationData
            Dict-like object containing the results.

        """
        if var_name not in self.PROVIDES_VARIABLES:
            raise ValueError(f"Invalid input variable {var_name}")

        # there's only one variable in the file
        aerocom_var_name = var_name

        # Iterate over the lines of the file
        self.logger.info(f"Reading file {filename}")
        file_delimiter = self.FILE_COL_DELIM
        # this lists the data to keep from the original read string
        # this becomes a time series
        file_indexes_to_keep = [11, 13, 14, 15, 16]
        # used for line length control...
        max_file_index_to_keep = max(file_indexes_to_keep)
        # this is some header information
        header_indexes_to_keep = [
            0,
            3,
            8,
            9,
            10,
            12,
        ]
        # These are the indexes with a time and are stored as np.datetime64
        time_indexes = [13, 14]

        # read the file
        # enable alternative reading of .gz files here to save space on the file system
        suffix = pathlib.Path(filename).suffix
        if suffix == ".gz":
            f_out = tempfile.NamedTemporaryFile(delete=False)
            with gzip.open(filename, "r") as f_in:
                shutil.copyfileobj(f_in, f_out)
            read_filename = f_out.name
            f_out.close()
        else:
            read_filename = filename

        # input files can be either UTF-8 or UTF-16 encoded
        # try both
        # files are max 3MB in size, so no big deal terms of RAM usage
        try:
            with open(read_filename) as f:
                lines = f.readlines()
        except UnicodeDecodeError:
            with open(read_filename, encoding="UTF-16") as f:
                lines = f.readlines()
        except Exception:
            if suffix == ".gz":
                os.remove(f_out.name)
            raise EEAv2FileError(f"Found corrupt file {filename}. consider deleteing it")

        # remove the temp file in case the input file was a gz file
        if suffix == ".gz":
            os.remove(f_out.name)

        header = lines[0].lower().rstrip().split(file_delimiter)
        # create output dict
        if len(header) < max_file_index_to_keep:
            raise EEAv2FileError(f"Found corrupt file {filename}. consider deleting it")

        data_dict = {}
        for idx in header_indexes_to_keep:
            data_dict[header[idx]] = ""

        for idx in file_indexes_to_keep:
            if idx in time_indexes:
                data_dict[header[idx]] = np.zeros(self.MAX_LINES_TO_READ, dtype="datetime64[s]")
            else:
                data_dict[header[idx]] = np.empty(self.MAX_LINES_TO_READ, dtype=np.float64)

        # read the data...
        # DE,http://gdi.uba.de/arcgis/rest/services/inspire/DE.UBA.AQD,NET.DE_BB,STA.DE_DEBB054,DEBB054,SPO.DE_DEBB054_PM2_dataGroup1,SPP.DE_DEBB054_PM2_automatic_light-scat_Duration-30minute,SAM.DE_DEBB054_2,PM2.5,http://dd.eionet.europa.eu/vocabulary/aq/pollutant/6001,hour,3.2000000000,µg/m3,2020-01-04 00:00:00 +01:00,2020-01-04 01:00:00 +01:00,1,2
        lineidx = 0
        for line in lines[1:]:
            rows = line.rstrip().split(file_delimiter)
            # Unfortunately there's a lot of corrupt files
            # skip data line if the # rows is not sufficient
            if len(rows) < max_file_index_to_keep:
                continue
            if lineidx == 0:
                for idx in header_indexes_to_keep:
                    if header[idx] != self.VAR_CODE_NAME:
                        data_dict[header[idx]] = rows[idx]
                    else:
                        # extract the EEA var code from the URL noted in the data file
                        data_dict[header[idx]] = rows[idx].split("/")[-1]

            for idx in file_indexes_to_keep:
                # if the data is a time
                if idx in time_indexes:
                    # make the time string ISO compliant so that numpy can directly read it
                    # this is not very time string forgiving but fast
                    data_dict[header[idx]][lineidx] = np.datetime64(
                        rows[idx][0:10] + "T" + rows[idx][11:19]
                    )
                    # due to the deprecation of the timezone interpretation after numpy 0.11
                    # we have to substract the offset manually to get to UTC.
                    # np.timedelta64 does not accept a float as parameter, only an integer.
                    # Although there are time zones with a 30 minutes offset, these don't
                    # exist in Europe, so just consider integer hours here for speed
                    tz_offset = np.timedelta64(np.int64(rows[idx][20:23]), "h")
                    data_dict[header[idx]][lineidx] = data_dict[header[idx]][lineidx] - tz_offset
                else:
                    # data is not a time
                    # sometimes there's no value in the file. Set that to nan
                    try:
                        data_dict[header[idx]][lineidx] = np.float64(rows[idx])
                    except (ValueError, IndexError):
                        data_dict[header[idx]][lineidx] = np.nan

            lineidx += 1
        # if the first line in the file was empty
        if data_dict["unitofmeasurement"] == "":
            if rows[12] == "":
                raise EEAv2FileError(
                    f"Unit of Measurment could not be inferred from EEA file {filename}"
                )
            else:
                # with loss of generality get the unitofmeasurement from the last row column 12 (which should be a kept header)
                data_dict["unitofmeasurement"] = rows[12]

        unit_in_file = data_dict["unitofmeasurement"]
        # adjust the unit and apply conversion factor in case we read a variable noted in self.AUX_REQUIRES
        if var_name in self.AUX_REQUIRES:
            unit_in_file = self.CONV_UNIT[var_name]
            data_dict[self.VAR_NAMES_FILE[var_name]] = (
                data_dict[self.VAR_NAMES_FILE[var_name]] * self.CONV_FACTOR[var_name]
            )
        try:
            unit = self.VAR_UNITS_FILE[unit_in_file]
        except KeyError:
            # this will raise an Exception if cf_units cannot handle. In
            # which case the unit should be added in VAR_UNITS_FILE
            unit = str(cf_units.Unit(unit_in_file))

        # Empty data object (a dictionary with extended functionality)
        data_out = StationData()
        data_out.data_id = self.data_id
        data_out.dataset_name = self.DATASET_NAME
        data_out.station_id = data_dict["airqualitystation"]
        data_out.station_name = data_dict["airqualitystation"]
        data_out.filename = filename
        data_out.instrument_name = self.INSTRUMENT_NAME
        data_out.country_code = data_dict["countrycode"]
        freq = data_dict["averagingtime"]
        try:
            tstype = self.TS_TYPES_FILE[freq]
        except KeyError:
            raise TemporalResolutionError(
                f"Found invalid ts_type {freq}. Please register in class header "
                f"attr TS_TYPES_FILE"
            )

        data_out.ts_type = tstype
        # ToDo: check "variables" entry, it should not be needed anymore in UngriddedData
        data_out["variables"] = [aerocom_var_name]
        data_out["var_info"][aerocom_var_name] = {}
        data_out["var_info"][aerocom_var_name]["units"] = unit
        # TsType is
        # data_out['var_info'][aerocom_var_name]['ts_type'] = self.TS_TYPE

        # Sometimes the times in the data files are not ordered in time which causes problems when doing
        # time interpolations later on. Make sure that the data is ordered in time
        diff_unsorted = np.diff(data_dict[self.START_TIME_NAME])
        sort_flag = False
        # use a vectorised time_diff instead of scalar one as before
        time_diff = (
            data_dict[self.END_TIME_NAME][:lineidx] - data_dict[self.START_TIME_NAME][:lineidx]
        ) / 2.0

        # np.min needs an array and fails with ValueError when a scalar is supplied
        # this is the case for a single line file
        try:
            min_diff = np.min(diff_unsorted)
        except ValueError:
            min_diff = 0

        if min_diff < 0:
            # data needs to be sorted
            ordered_idx = np.argsort(data_dict[self.START_TIME_NAME][:lineidx])
            data_out["dtime"] = (
                data_dict[self.START_TIME_NAME][ordered_idx] + time_diff[ordered_idx]
            )
            sort_flag = True
        else:
            data_out["dtime"] = data_dict[self.START_TIME_NAME][:lineidx] + time_diff

        for key, value in data_dict.items():
            # adjust the variable name to aerocom standard
            if key != self.VAR_NAMES_FILE[aerocom_var_name]:
                data_out[key] = value[:lineidx]
            else:
                if sort_flag:
                    data_out[aerocom_var_name] = value[:lineidx][ordered_idx]
                else:
                    data_out[aerocom_var_name] = value[:lineidx]

        # convert data vectors to pandas.Series (if attribute
        # vars_as_series=True)
        if vars_as_series:
            data_out[aerocom_var_name] = pd.Series(
                data_out[aerocom_var_name], index=data_out["dtime"]
            )

        return data_out

    def _read_metadata_file(self, filename=None):
        """Read EEA metadata file

        Parameters
        ----------
        filename : str
            Absolute path to filename to read.

        Returns
        -------
        metadata
            Dict-like object containing the results with the keys being a combination of the
            station name and the variable and the values being all fields of the metadata file

        """

        if filename is None:
            filename = os.path.join(self.data_dir, self.DEFAULT_METADATA_FILE)
            # test also for a gzipped file...
        if not os.path.isfile(filename):
            filename = filename + ".gz"
        self.logger.warning(f"Reading file {filename}")

        struct_data = {}
        suffix = pathlib.Path(filename).suffix
        if suffix == ".gz":
            f_out = tempfile.NamedTemporaryFile(delete=False)
            with gzip.open(filename, "r") as f_in:
                shutil.copyfileobj(f_in, f_out)
            filename = f_out.name
            f_in.close()

        with open(filename) as f:
            # read header...
            # Countrycode Timezone Namespace   AirQualityNetwork AirQualityStation AirQualityStationEoICode   AirQualityStationNatCode   SamplingPoint  SamplingProces Sample   AirPollutantCode  ObservationDateBegin ObservationDateEnd   Projection  Longitude   Latitude Altitude MeasurementType   AirQualityStationType   AirQualityStationArea   EquivalenceDemonstrated MeasurementEquipment InletHeight BuildingDistance  KerbDistance
            header = f.readline().lower().rstrip().split()
            min_row_no = len(header)
            # create output dict
            data_dict = {}
            for key in header:
                data_dict[key] = []
            lineidx = 0
            bad_line_no = 0
            bad_line_arr = []
            for line in f:
                rows = line.rstrip().split("\t")
                # skip too short lines
                if len(rows) < min_row_no:
                    # print(line)
                    bad_line_no += 1
                    bad_line_arr.append(line)
                    continue
                temp_dict = {}
                for idx, key in enumerate(header):
                    if header[idx] != self.VAR_CODE_NAME:
                        # data_dict[header[idx]] = rows[idx]
                        temp_dict[header[idx]] = rows[idx]
                    else:
                        # extract the EEA var code from the URL noted in the data file
                        # data_dict[header[idx]] = rows[idx].split('/')[-1]
                        temp_dict[header[idx]] = rows[idx].split("/")[-1]

                meta_key = f"{temp_dict['airqualitystation']}__{temp_dict['airpollutantcode']}"
                if meta_key not in struct_data:
                    struct_data[meta_key] = temp_dict.copy()
                else:
                    pass
                lineidx += 1

        self.logger.info(f"Reading file {filename} done")
        # remove the temp file in case the input file was a gz file
        if suffix == ".gz":
            f_out.close()
            os.remove(f_out.name)

        return struct_data

    def get_file_list(self, pattern=None):
        """Search all files to be read

        Uses :attr:`_FILEMASK` (+ optional input search pattern, e.g.
        station_name) to find valid files for query.

        Parameters
        ----------
        pattern : str, optional
            file name pattern applied to search

        Returns
        -------
        list
            list containing retrieved file locations

        Raises
        ------
        IOError
            if no files can be found
        """

        if pattern is None:
            logger.warning("using default pattern *.* for file search")
            pattern = "*.*"

        self.logger.info("Fetching data files. This might take a while...")
        fp = os.path.join(self.data_dir, pattern)
        return sorted(glob.glob(fp, recursive=True))

    def get_station_coords(self, meta_key):
        """
        get a station's coordinates

        Parameters
        ----------
        meta_key : str
            string with the internal station key
        """
        ret_data = {}
        ret_data["latitude"] = float(self._metadata[meta_key][self.LATITUDENAME])
        ret_data["longitude"] = float(self._metadata[meta_key][self.LONGITUDENAME])
        ret_data["altitude"] = float(self._metadata[meta_key][self.ALTITUDENAME])
        return ret_data

    def read(
        self, vars_to_retrieve=None, files=None, first_file=None, last_file=None, metadatafile=None
    ):
        """Method that reads list of files as instance of :class:`UngriddedData`

        Parameters
        ----------
        vars_to_retrieve : :obj:`list` or similar, optional
            List containing variable IDs that are supposed to be read. If None,
            all variables in :attr:`PROVIDES_VARIABLES` are loaded.
        files : :obj:`list`, optional
            List of files to be read. If None, then the file list used is the
            returned from :func:`get_file_list`.
        first_file : :obj:`int`, optional
            Index of the first file in :obj:'file' to be read. If None, the
            very first file in the list is used.
        last_file : :obj:`int`, optional
            Index of the last file in :obj:'file' to be read. If None, the very
            last file in the list is used.
        metadatafile : :obj:'str', optional
            full qualified path to metadata file. If None, the default metadata
            file will be used

        Returns
        -------
        UngriddedData
            data object
        """

        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]

        if len(vars_to_retrieve) > 1:
            raise NotImplementedError("So far, only one variable can be read at a time...")
        var_name = vars_to_retrieve[0]
        logger.info("Reading EEA data")
        if files is None:
            logger.info("Retrieving file list")
            files = self.get_file_list(self.FILE_MASKS[var_name])

        if first_file is None:
            first_file = 0
        if last_file is None:
            last_file = len(files)

        if metadatafile is None:
            metadatafile = os.path.join(self.data_dir, self.DEFAULT_METADATA_FILE)

        files = files[first_file:last_file]

        data_obj = UngriddedData()
        meta_key = 0.0
        idx = 0

        # Assign metadata object and index
        metadata = data_obj.metadata
        meta_idx = data_obj.meta_idx

        logger.info("Reading metadata file")
        # non compliant, but efficiently indexed metadata
        self._metadata = self._read_metadata_file(metadatafile)

        # returns a dict with country codes as keys and the country names as value
        _country_dict = get_country_name_from_iso()
        logger.info("Reading files...")

        for i in tqdm(range(len(files)), disable=None):
            _file = files[i]
            try:
                station_data = self.read_file(_file, var_name=var_name)
            except EEAv2FileError:
                self.logger.warning(f"file {_file} is corrupt! consider deleting it")
                continue
            except TemporalResolutionError as e:
                self.logger.warning(f"{_file} has TemporalResolutionError")
                logger.warning(f"{repr(e)}. Skipping file...")
                continue

            # readfile might fail outside of the error captured by the try statement above
            if station_data is None:
                self.logger.warning(f"file {_file} did not provide data. skipping...!")
                continue

            # to find the metadata quickly, we use a string internally
            _meta_key = f"{station_data['station_id']}__{station_data['airpollutantcode']}"

            # Fill the metadata dict.
            # The location in the data set is time step dependant
            if _meta_key not in self._metadata:
                self.logger.warning(
                    f"metadata for station {_meta_key} not found! skipping that station!"
                )
                continue
            metadata[meta_key] = {}
            meta_idx[meta_key] = {}
            metadata[meta_key].update(station_data.get_meta())
            metadata[meta_key].update(self.get_station_coords(_meta_key))
            metadata[meta_key]["variables"] = list(
                station_data.var_info.keys()
            )  # vars_to_retrieve
            metadata[meta_key]["station_classification"] = self._metadata[_meta_key][
                "airqualitystationtype"
            ]
            metadata[meta_key]["area_classification"] = self._metadata[_meta_key][
                "airqualitystationarea"
            ]
            try:
                metadata[meta_key]["country"] = _country_dict[metadata[meta_key]["country_code"]]
            except KeyError:
                pass
            metadata[meta_key]["var_info"] = station_data["var_info"]
            metadata[meta_key]["website"] = self.WEBSITE
            metadata[meta_key]["data_product"] = self.DATA_PRODUCT
            metadata[meta_key]["station_name"] = self._metadata[_meta_key][
                "airqualitystationeoicode"
            ]

            # List with indices of this station for each variable
            num_times = len(station_data["dtime"])

            # Check whether the size of the data object needs to be extended
            if (idx + num_times) >= data_obj._ROWNO:
                # if totnum < data_obj._CHUNKSIZE, then the latter is used
                data_obj.add_chunk(num_times)

            for var_idx, var in enumerate(list(station_data.var_info)):
                # set invalid data to np.nan according to
                # https://dd.eionet.europa.eu/vocabulary/aq/observationvalidity/view
                # data flagged as below the detection limit (values 2 and 3)
                # will remain in the data
                station_data[var][station_data["validity"] < 1] = np.nan

                values = station_data[var]
                start = idx + var_idx * num_times
                stop = start + num_times

                data_obj._data[start:stop, data_obj._METADATAKEYINDEX] = meta_key
                data_obj._data[start:stop, data_obj._TIMEINDEX] = station_data["dtime"]
                data_obj._data[start:stop, data_obj._DATAINDEX] = values
                data_obj._data[start:stop, data_obj._VARINDEX] = var_idx
                meta_idx[meta_key][var] = np.arange(start, stop)

                if var not in data_obj.var_idx:
                    data_obj.var_idx[var] = var_idx

            idx += num_times
            meta_key = meta_key + 1.0

        # Shorten data_obj._data to the right number of points
        data_obj._data = data_obj._data[:idx]
        # data_obj.data_revision[self.DATASET_NAME] = self.data_revision
        self._metadata = None
        self.files = files

        return data_obj
